
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <math.h>

#include "structs.h"
#include "utils.h"
#include "enum.h"
#include "sample.h"
#include "induced.h"
#include "mersenne.h"

TreeNode * rootNode;
unsigned long numLeaves;
unsigned long sampleSize, numSamples;
double minDist, maxDist;
double numValidSamples;

#ifdef _MEM_INFO_
double CurMemoryUse, MaxMemoryUse;
#endif

// Just prints out how the program should be run and exits.
void Depth_ShowUsage(char * command)
{
  printf("Usage: %s (options) <tree_file> (<sample_size> <num_samples> <min_dist> <max_dist>)\n\n", command);
  printf("options include:\n");
  printf("\t-q\tquiet mode, samples will not be displayed\n");
  printf("\t-s\tuse short taxon names (truncate to 20 characters)\n");
  printf("\t-v\tverify that the sample satisfies the constraints\n");
  printf("\t-t\toutput Newick tree instead of sample list\n");
  printf("\t-b\toutput Newick tree and sample list\n");
  printf("\t-i\tdisplay some basic summary info\n");
  exit (-1);
}

int main(int argc, char * argv[])
{
 FILE *fp;
  unsigned long i;
  TreeNodeList * Sample, * curNode;
  InducedSubtree * subtree;
  int quiet = 0;
  int onlyTree = 0;
  int treeAndList = 0;
  int info = 0;
  int verify = 0;
  int shortNames = 0;

#ifdef _MEM_INFO_
  CurMemoryUse = 0.0;
  MaxMemoryUse = 0.0;
#endif

  // Set the randomizer seed
  Depth_sgenrand(time(0));

  // Check for options first
  for (i=1; i < argc; i++) {
    if (argv[i][0] == '-') {
      if (argv[i][1] == 'q')
        quiet = 1;
      else if (argv[i][1] == 't')
        onlyTree = 1;
      else if (argv[i][1] == 'b')
        treeAndList = 1;
      else if (argv[i][1] == 'i')
        info = 1;
      else if (argv[i][1] == 'v')
        verify = 1;
      else if (argv[i][1] == 's')
        shortNames = 1;
      else
        Depth_ShowUsage(argv[0]);
    } else {
      break;
    }
  }



/* 

  // They didn't include a file name
*/
  if (i == argc)
    Depth_ShowUsage(argv[0]);

  // Set some default values
  numSamples = 1;
  sampleSize = 2;
  minDist = 0.0;
  maxDist = 20.0;

  // Allow the user to specify other options
  if (argc > i+1)
    sampleSize = atoi(argv[i+1]);
 
  ASSERT(sampleSize >= 2, "You must select a sample size of at least 2.")
 
  if (argc > i+2)
    numSamples = atoi(argv[i+2]);
 
  if (argc > i+3)
    minDist = atof(argv[i+3]);
  if (argc >= i+4)
    maxDist = atof(argv[i+4]);
 
  ASSERT(minDist < maxDist, "You must select a maximum distance that is larger than the minimum distance.");


  // Open up the tree file
  fp = fopen(argv[i], "r");
  ASSERT2(fp, "Could not open file '%s'.\n", argv[i]);

  // This will read in the tree so that the global
  // variable "rootNode" points to the root
  rootNode = Depth_ReadInTree(fp, shortNames);
  fclose(fp);
  ASSERT(rootNode, "Unexpected error reading file.");

  // This will traverse the tree, counting samples
  // rooted at each node
  numValidSamples = Depth_CountSamples(rootNode, 0.0);
  ASSERT(numValidSamples > 0.0, "Sorry, but there are no valid samples satisfying the given constraints.");

  // Start picking some samples uniformly at random
  for (i=0; i < numSamples; i++) {

    // This will choose and print out a sample of the correct size
    Sample = Depth_SelectSample();
    ASSERT2(Sample, "Unexpected error generating sample %d.", i+1);

    if (!quiet) {
      printf("Sample %d:\n", i+1);
 
      // Print out the leaves in the sample
      if (!onlyTree) {
        curNode = Sample;
        while (curNode) {
          printf("  %s\n", curNode->Node->Name);
          curNode = curNode->Next;
        }
      }

      if (verify || onlyTree || treeAndList) {
 
        InducedSubtree * Subtree;
 
        // Construct the subtree induced by the sample and verify
        // that it satisfies the given constraints
        Subtree = Depth_BuildInducedSubtree(Sample);
	ASSERT2(Subtree, "Unexpected error trying to build the subtree induced by sample %d.", i+1);
	if (verify)
	  Depth_CheckInducedSubtree(Subtree);

        // If the user wants the subtree printed out in Newick
        // format, do it now
        if (onlyTree || treeAndList)
          Depth_PrintInducedSubtree(Subtree, 0);
 
        Depth_DeleteInducedSubtree(Subtree);
      }
    }
 
    Depth_DeleteList(Sample);
  }

  // Print some summary data, if the user wants it
  if (info) {
    printf("Number of leaves: %d\n", numLeaves);
    printf("Valid samples: %e (out of %e possible samples)\n", numValidSamples,
           Depth_fchoose((double)numLeaves, (double)sampleSize));

#ifdef _MEM_INFO_
    printf("Memory use: %.0f bytes\n", MaxMemoryUse);
#endif 
   }

  // Clean up
  Depth_DeleteTree(rootNode);
  
  return 0;
}
