
#include <stdlib.h>

#include "structs.h"
#include "utils.h"

extern double treeListNodes;
extern double maxDist, minDist;
extern unsigned long sampleSize;

#ifdef _MEM_INFO_
extern double CurMemoryUse, MaxMemoryUse;
#endif

// Passes a list of descendents up the tree in order to calculate
// the number of samples rooted at each internal node. Returns
// the number of samples rooted at this node or below.
double Depth_CountSamples(TreeNode * node, double curDepth)
{
  double curValidSamples = 0.0;
  TreeNodeList * LeftList, * RightList;
  TreeNodeList * curDesc = NULL;
  unsigned long numLeftValid = 0, numRightValid = 0;
  int fromLeft;

  // Update the depth of a node. Initially the depth of a node
  // was set to be the distance from its parent. We want the
  // depth to be the distance to the root, so fix this now.
  node->depth += curDepth;

  // If it is a leaf then just add it to its own descendent list
  if (isLeaf(node)) {
    node->Descendents = (TreeNodeList *)malloc(sizeof(TreeNodeList));
    ASSERT(node->Descendents, "Out of memory.");
#ifdef _MEM_INFO_
    INC_MEM(sizeof(TreeNodeList));
#endif
    node->Descendents->Node = node;
    node->Descendents->Next = NULL;
    return 0.0;
  }

  // Calculate the number of samples rooted below here
  curValidSamples = Depth_CountSamples(node->LChild, node->depth);
  curValidSamples += Depth_CountSamples(node->RChild, node->depth);

  // Merge the descendent lists from the two child nodes. Destroys
  // the descendent lists from the two children.
  LeftList = node->LChild->Descendents;
  RightList = node->RChild->Descendents;
  node->Descendents = NULL;
  for (;;) {
    
    if (!LeftList && !RightList) {
      curDesc->Next = NULL;
      break;
    }

    // Add a node off the front of the left or right child s
    // descendent list (the one with the shallowest depth)
    if (!LeftList) {
      fromLeft = 0;
    } else if (!RightList) {
      fromLeft = 1;
    } else if (LeftList->Node->depth <= RightList->Node->depth) {
      fromLeft = 1;
    } else {
      fromLeft = 0;
    }

    // Add the node to our list and move to the next node in
    // the childs descendent list
    if (fromLeft) {

      // Stop when the leaves are too far away
      if (LeftList->Node->depth - node->depth >= maxDist) {
	curDesc->Next = NULL;
	Depth_DeleteList(LeftList);
	Depth_DeleteList(RightList);
	break;
      }

      if (curDesc) {
	curDesc->Next = LeftList;
	curDesc = curDesc->Next;
      } else {
	node->Descendents = LeftList;
	curDesc = LeftList;
      }

      LeftList = LeftList->Next;

    } else {

      // Stop when the leaves are too far away
      if (RightList->Node->depth - node->depth >= maxDist) {
	curDesc->Next = NULL;
	Depth_DeleteList(LeftList);
	Depth_DeleteList(RightList);
	break;
      }

      if (curDesc) {
	curDesc->Next = RightList;
	curDesc = curDesc->Next;
      } else {
	node->Descendents = RightList;
	curDesc = RightList;
      }

      RightList = RightList->Next;

    }

    // Count the number of leaves that are within the required
    // distance from this node
    if (curDesc->Node->depth - node->depth > minDist) {
      if (fromLeft) {
	numLeftValid++;
      } else {
	numRightValid++;
      }
    }
  }

  // If there is at least one valid sample rooted here, update
  // the number of current samples
  if (numLeftValid > 0 && numRightValid > 0 && 
      numLeftValid + numRightValid >= sampleSize) {
    
    node->SamplesRootedHere = Depth_fchoose((double)(numLeftValid+numRightValid), (double)sampleSize);

    // Remove samples that are taken completely from the left
    // subtree or completely from the right
    node->SamplesRootedHere -= Depth_fchoose((double)numLeftValid, (double)sampleSize);
    node->SamplesRootedHere -= Depth_fchoose((double)numRightValid, (double)sampleSize);
    
    curValidSamples += node->SamplesRootedHere;
  } else {
    node->SamplesRootedHere = 0.0;
  }

  node->numValidLDescendents = numLeftValid;
  node->numValidRDescendents = numRightValid;

  // The descendents of the children are no longer needed (and we have
  // messed up their lists anyway)
  node->LChild->Descendents = NULL;
  node->RChild->Descendents = NULL;

  return curValidSamples;
}
