
#include <stdlib.h>

#include "structs.h"
#include "utils.h"

extern double minDist, maxDist;
extern unsigned long sampleSize;

#ifdef _MEM_INFO_
extern double CurMemoryUse, MaxMemoryUse;
#endif

// Find the least common ancestor of two nodes. 
//
// TODO: Our method here is really dumb and takes up to O(h) time 
//       (where h is the height of the tree), in the worst case. We should 
//       implement this to run in O(1) time.
TreeNode * Depth_LCA(TreeNode * node1, TreeNode * node2)
{
  ASSERT(node1 && node2, "Depth_LCA was called with a NULL node!");

  for (;;) {
    if (node1 == node2)
      return node1;

    if (node1->depth < node2->depth)
      node2 = node2->Parent;
    else
      node1 = node1->Parent;
  }  
}

// Recursively deletes the subtree rooted here
void Depth_DeleteInducedSubtree(InducedSubtree * node)
{
  if (node) {
    Depth_DeleteInducedSubtree(node->LChild);
    Depth_DeleteInducedSubtree(node->RChild);
    free(node);
#ifdef _MEM_INFO_
    DEC_MEM(sizeof(InducedSubtree);
#endif
  }
}

// Creates a subtree wrapper node for the give tree node.
InducedSubtree * Depth_CreateInducedNode(TreeNode * node)
{
  InducedSubtree * newNode;
  newNode = (InducedSubtree *)malloc(sizeof(InducedSubtree));
  ASSERT(newNode, "Out of memory.");
#ifdef _MEM_INFO_
    INC_MEM(sizeof(InducedSubtree);
#endif
  newNode->Node = node;
  newNode->Parent = NULL;
  newNode->LChild = NULL;
  newNode->RChild = NULL;
  return newNode;
}

// This is really easy to do because we know that the samples we
// get are ordered left-to-right, as in the original tree. We can
// just keep adding one node at a time by finding the least common
// ancestor with the right-most node in the induced subtree and
// the next node coming in.
//
// TODO: We could check that the depth constraints are not
//       violated as the induced subtree is being built instead
//       of traversing the tree afterwards to check this.
InducedSubtree * Depth_BuildInducedSubtree(TreeNodeList * nodes)
{
  InducedSubtree * subtree, * prevNode, * newNode, * lcaSubNode;
  TreeNode * lca;

  ASSERT(nodes && nodes->Next, "Sample contains less than two nodes!");

  // Get the first two nodes off the list and find their LCA
  prevNode = Depth_CreateInducedNode(nodes->Node);
  nodes = nodes->Next;
  newNode = Depth_CreateInducedNode(nodes->Node);
  nodes = nodes->Next;
  subtree = Depth_CreateInducedNode(Depth_LCA(prevNode->Node, newNode->Node));
  subtree->LChild = prevNode;
  subtree->RChild = newNode;
  prevNode->Parent = subtree;
  newNode->Parent = subtree;

  while (nodes) {

    // Pull the next node off the list to be added
    prevNode = newNode;
    newNode = Depth_CreateInducedNode(nodes->Node);
    nodes = nodes->Next;

    // Find the LCA (in the original tree) of this new node 
    // with the right-most node in the induced subtree
    lca = Depth_LCA(prevNode->Node, newNode->Node);

    // Walk up the right branch of the tree until we either get
    // past the root of the induced subtree, or we find the
    // correct spot along the branch to insert the LCA node.
    for (;;) {
      if (!prevNode || prevNode->Node->depth <= lca->depth)
	break;
      prevNode = prevNode->Parent;
    }

    lcaSubNode = Depth_CreateInducedNode(lca);
    if (!prevNode) {

      // New root for the induced subtree
      lcaSubNode->LChild = subtree;
      lcaSubNode->RChild = newNode;
      newNode->Parent = lcaSubNode;
      subtree->Parent = lcaSubNode;
      subtree = lcaSubNode;
    } else {

      // We should never end up with an LCA node thats already
      // in the tree. If we do, something went wrong.
      ASSERT(prevNode->Node != lca, "Error building induced subtree!");

      lcaSubNode->LChild = prevNode->RChild;
      lcaSubNode->RChild = newNode;
      lcaSubNode->Parent = prevNode;
      prevNode->RChild = lcaSubNode;
      lcaSubNode->LChild->Parent = lcaSubNode;
      newNode->Parent = lcaSubNode;      
    }
  }

  return subtree;
}

// Checks that no leaf in the sample is too close or too far
// from the root of the sample. Returns the number of leaves
// in the sample.
unsigned long Depth_CheckSubtreeHelper(InducedSubtree * node, double minDepth,
				       double maxDepth)
{
  ASSERT(node, "CheckSubtreeHelper() called with a NULL node.");

  if (isLeaf(node->Node)) {
    ASSERT(node->Node->depth > minDepth, "A leaf in the sample was too shallow!\n");
    ASSERT(node->Node->depth < maxDepth, "A leaf in the sample was too deep!\n");
    return 1;
  } else {
    return Depth_CheckSubtreeHelper(node->LChild, minDepth, maxDepth)
         + Depth_CheckSubtreeHelper(node->RChild, minDepth, maxDepth);
  }
}

// Calls the helper function described above making sure that
// the minimum and maximum distances are relative to the root of the
// induced subtree, rather than the entire tree.
void Depth_CheckInducedSubtree(InducedSubtree * node)
{
  unsigned long leaves;
  leaves = Depth_CheckSubtreeHelper(node, node->Node->depth + minDist,
				    node->Node->depth + maxDist);

  ASSERT(leaves == sampleSize, "The chosen sample is not the right size!");
}

// Prints out the given subtree in Newick format. External calls
// to PrintSubtree should set indent to 0.
void Depth_PrintInducedSubtree(InducedSubtree * node, int indent)
{
  int i, k;

  // Add some spacing to show hierarchy
  for (k=0; k < indent; k++)
    printf(" ");

  if (isLeaf(node->Node)) {
    double dist = node->Node->depth - node->Parent->Node->depth;

    printf("\'");
    for (i=0; node->Node->Name[i] != '\0'; i++) {
      char c = node->Node->Name[i];
      if (c == '[' || c == '(')
	c = '<';
      else if (c == ']' || c == ')')
	c = '>';
      else if (c == ';' || c == ':')
	continue;

      printf("%c", c);
    }
	
    printf("\'");
    if (dist < 0.0001)
      printf(":%e", dist);
    else
      printf(":%f", dist);
  } else {

    // Recursively print out child subtrees
    printf("(\n");
    Depth_PrintInducedSubtree(node->LChild, indent+1);
    printf(",\n");
    Depth_PrintInducedSubtree(node->RChild, indent+1);
    printf("\n");

    // Indent accordingly
    for (k=0; k < indent; k++)
      printf(" ");

    if (!node->Parent) {
      printf(");\n");
    } else {
      double dist = node->Node->depth - node->Parent->Node->depth;
      if (dist < 0.0001)
	printf("):%e", dist);
      else
	printf("):%f", dist);
    }
  }
}
