
#include <stdlib.h>
#include <math.h>

#include "structs.h"
#include "utils.h"
#include "mersenne.h"

extern TreeNode * rootNode;
extern double numValidSamples;
extern unsigned long sampleSize;
extern double minDist, maxDist;

#ifdef _MEM_INFO_
extern double CurMemoryUse, MaxMemoryUse;
#endif

// Does an inorder traversal of the tree, finding the correct sample
// root based on the point at which the sum of the samples adds up to
// the given rootIndex (which is between 0 and numValidSamples).
//
// If the root is found, "samRoot" will point to it and -1.0 is
// returned. If not, it will return the current index, after
// subtracting the number of valid samples rooted at each of
// the nodes that have been visited.
double Depth_FindSampleRoot(TreeNode * node, double rootIndex, 
			    TreeNode ** samRoot)
{
  double curIndex;

  ASSERT(node && samRoot, "Invalid paramter to FindSampleRoot().");

  // The root of the induced subtree cant be a leaf
  if (isLeaf(node))
    return rootIndex;

  // Try left subtree
  curIndex = Depth_FindSampleRoot(node->LChild, rootIndex, samRoot);
  if (curIndex == -1.0)
    return -1.0;
  
  // Try this node
  if (curIndex <= node->SamplesRootedHere) {
    *samRoot = node;
    return -1.0;
  }
  curIndex -= node->SamplesRootedHere;

  // Try right subtree
  return Depth_FindSampleRoot(node->RChild, curIndex, samRoot);
}

// Given an array of 0/1 values (called index), where index[i]=1 means
// that the (i-1)th valid leaf should be included in the sample, and
// 0 means that it shouldnt, this function traverses the subtree rooted 
// at the given node, returning the list of valid leaves represented
// by the index array.
TreeNodeList * Depth_FindLeavesHelper(TreeNode * node, double minDepth,
				      double maxDepth, int * NodeVector,
				      int * index)
{
  TreeNodeList * list, * llist, * rlist;

  if (isLeaf(node)) {
    
    // If this is a valid leaf for the sample, check the NodeVector
    // to see if its supposed to be chosen
    if (node->depth > minDepth && node->depth < maxDepth) {
      if (NodeVector[*index]) {
	list = (TreeNodeList *)malloc(sizeof(TreeNodeList));
#ifdef _MEM_INFO_
	INC_MEM(sizeof(TreeNodeList));
#endif
	list->Node = node;
	list->Next = NULL;
      } else {
	list = NULL;
      }
      (*index) = (*index) + 1;
      return list;
    }
    return NULL;
  }

  // Recursively try the left child
  llist = Depth_FindLeavesHelper(node->LChild, minDepth, maxDepth, NodeVector,
				 index);

  // Now, try the right child (if there were some leaves returned from
  // the left child, make sure the leaves from the right child get appended 
  // on to the end of the list)
  if (llist) {
    list = llist;
    while (llist->Next)
      llist = llist->Next;
    llist->Next = Depth_FindLeavesHelper(node->RChild, minDepth, maxDepth,
					 NodeVector, index);
  } else {
    list = Depth_FindLeavesHelper(node->RChild, minDepth, maxDepth, NodeVector,
				  index);
  }

  return list;
}

// Just calls the helper function described above
TreeNodeList * Depth_FindLeaves(TreeNode * node, int * NodeVector,
				double minLeafDepth, double maxLeafDepth)
{
  int index = 0;

  return Depth_FindLeavesHelper(node, minLeafDepth, maxLeafDepth, 
				NodeVector, &index);
}

// Selects one valid sample uniformly at random from the tree. Assumes
// that the valid samples have already been enumerated at each node.
//
// TODO: The method of randomly selecting positions for 1s in LVector
//       and RVector is quite inefficient.
TreeNodeList * Depth_SelectSample()
{
  TreeNode * sampleRoot = NULL;
  double rootIndex;
  unsigned long ldesc, rdesc, i, j, index, sum, lleft, rleft, realIndex;
  int * LVector, * RVector;
  TreeNodeList * Sample, * temp;

  //
  // Pick a root for the sample uniformly at random
  //
  rootIndex = numValidSamples * Depth_genrand();
  Depth_FindSampleRoot(rootNode, rootIndex, &sampleRoot);
  ASSERT(sampleRoot, "Failed to choose a root for this sample.");

  //
  // Decide which leaves to pick for the sample
  //
  ldesc = sampleRoot->numValidLDescendents;
  rdesc = sampleRoot->numValidRDescendents;

  LVector = (int *)malloc(ldesc * sizeof(int));
  ASSERT(LVector, "Out of memory.");
  RVector = (int *)malloc(rdesc * sizeof(int));
  ASSERT(RVector, "Out of memory.");
#ifdef _MEM_INFO_
  INC_MEM((rdesc+ldesc) * sizeof(int));
#endif

  // Initialize all leaves to unpicked
  for (i=0; i < ldesc; i++)
    LVector[i] = 0;
  for (i=0; i < rdesc; i++)
    RVector[i] = 0;

  // Pick one leaf from left and right subtrees
  index = floor(Depth_genrand() * (double)ldesc);
  if (index == ldesc)
    index -= 1;
  LVector[index] = 1;
  index = floor(Depth_genrand() * (double)rdesc);
  if (index == rdesc)
    index -= 1;
  RVector[index] = 1;

  // Start picking the rest of the sample
  lleft = ldesc-1;
  rleft = rdesc-1;
  for (i=2; i < sampleSize; i++) {
    index = floor(Depth_genrand() * (double)(lleft+rleft));
    if (index == lleft+rleft)
      index--;

    // Find the correct leaf (we need to ignore nodes which
    // are already set to 1, since we only picked from among
    // the unselected nodes)
    if (index < lleft) {
      realIndex = 0;
      for (j=0;; j++) {
	if (LVector[j] == 0) {
	  if (realIndex == index)
	    break;
	  else
	    realIndex++;
	}
      }

      LVector[j] = 1;
      lleft--;
    } else {
      realIndex = 0;
      for (j=0;; j++) {
	if (RVector[j] == 0) {
	  if (realIndex == index - lleft)
	    break;
	  else
	    realIndex++;
        }
      }

      RVector[j] = 1;
      rleft--;
    }
  }

  // Find the leaves
  Sample = Depth_FindLeaves(sampleRoot->LChild, LVector,
			    sampleRoot->depth + minDist,
			    sampleRoot->depth + maxDist);
  ASSERT(Sample, "Failed to select leaves for this sample.");

  temp = Sample;
  while (temp->Next) {
    temp = temp->Next;
  }

  temp->Next = Depth_FindLeaves(sampleRoot->RChild, RVector,
				sampleRoot->depth + minDist,
				sampleRoot->depth + maxDist);
  ASSERT(temp->Next, "Failed to select leaves for this sample.");

  free(LVector);
  free(RVector);
#ifdef _MEM_INFO_
  DEC_MEM((rdesc+ldesc) * sizeof(int));
#endif

  // Return the sample
  return Sample;
}

