
#include <stdlib.h>
 
#include "structs.h"
#include "utils.h"

extern unsigned long sampleSize;
extern double minDist, maxDist;

#ifdef _MEM_INFO_
extern double CurMemoryUse, MaxMemoryUse;
#endif

// Passes a list of descendents up the tree in order to calculate
// the number of samples rooted at each internal node. Returns
// the number of samples rooted at this node or below of the
// desired size.
//
// TODO: Leaves do not really need a SamplesRootedHere array since
//       they will always have a 1 in the first entry and 0 everywhere
//       else. Half of the arrays can be eliminated by getting rid
//       of these.
// TODO: For any node with fewer than j leaf descendents, entries
//       with index >= j will always be zero, so we can cut these
//       out too (although the savings in space may not be worth
//       the extra complexity).
double Edge_CountAllSamples(TreeNode * node, double curDepth)
{
  double curValidSamples = 0.0;
  TreeNodeList * LeftList, * RightList;
  TreeNodeList * curDesc = NULL;
  double * FromLeft, * FromRight;
  unsigned long fromLeft;
  unsigned long i,j;

  // Update the depth of a node. Initially the depth of a node
  // was set to be the distance from its parent. We want the
  // depth to be the distance to the root, so fix this now.
  node->depth = curDepth + node->depth;

  // Need an array to store the number of samples of each size
  // rooted at this internal node
  node->SamplesRootedHere = (double *)malloc(sampleSize * sizeof(double));
  ASSERT(node->SamplesRootedHere, "Out of memory.");
#ifdef _MEM_INFO_
  INC_MEM(sampleSize * sizeof(double));
#endif

  if (isLeaf(node)) {
    node->Descendents = (TreeNodeList *)malloc(sizeof(TreeNodeList));
    ASSERT(node->Descendents, "Out of memory.");
#ifdef _MEM_INFO_
    INC_MEM(sizeof(TreeNodeList));
#endif

    node->Descendents->Node = node;
    node->Descendents->Next = NULL;

    // A leaf just has one sample of size 1 rooted here and
    // all other entries should be 0
    node->SamplesRootedHere[0] = 1.0;
    for (i=1; i < sampleSize; i++)
      node->SamplesRootedHere[i] = 0.0;

    return 0.0;
  }

  // Need arrays to store samples of each size from left and right subtrees
  FromLeft = (double *)malloc((sampleSize-1) * sizeof(double));
  ASSERT(FromLeft, "Out of memory.");
  FromRight = (double *)malloc((sampleSize-1) * sizeof(double));
  ASSERT(FromRight, "Out of memory.");
#ifdef _MEM_INFO_
  INC_MEM(2 * (sampleSize-1) * sizeof(double));
#endif

  // Initialize all the entries
  for (i=0; i < sampleSize-1; i++) {
    node->SamplesRootedHere[i] = 0.0;
    FromLeft[i] = 0.0;
    FromRight[i] = 0.0;
  }
  node->SamplesRootedHere[sampleSize-1] = 0.0;

  // Calculate the number of valid samples rooted below here
  curValidSamples = Edge_CountAllSamples(node->LChild, node->depth);
  curValidSamples += Edge_CountAllSamples(node->RChild, node->depth);

  // This node will be the first in its own descendent list
  node->Descendents = (TreeNodeList *)malloc(sizeof(TreeNodeList));
  ASSERT(node->Descendents, "Out of memory.");
#ifdef _MEM_INFO_
  INC_MEM(sizeof(TreeNodeList));
#endif

  node->Descendents->Node = node;
  curDesc = node->Descendents;

  // Merge the descendent lists from the two child nodes
  LeftList = node->LChild->Descendents;
  RightList = node->RChild->Descendents;
  for (;;) {
    
    if (!LeftList && !RightList) {
      curDesc->Next = NULL;
      break;
    }

    // Add a node off the front of the left or right childs
    // descendent list (the one with the shallowest depth)
    if (!LeftList) {
      fromLeft = 0;
    } else if (!RightList) {
      fromLeft = 1;
    } else if (LeftList->Node->depth <= RightList->Node->depth) {
      fromLeft = 1;
    } else {
      fromLeft = 0;
    }

    if (fromLeft) {

      // Stop when the leaves are too far away
      if (LeftList->Node->depth - node->depth >= maxDist) {
	curDesc->Next = NULL;
	Edge_DeleteList(LeftList);
	Edge_DeleteList(RightList);
	break;
      }

      curDesc->Next = LeftList;
      curDesc = curDesc->Next;
      LeftList = LeftList->Next;
    } else {

      // Stop when the leaves are too far away
      if (RightList->Node->depth - node->depth >= maxDist) {
	curDesc->Next = NULL;
	Edge_DeleteList(LeftList);
	Edge_DeleteList(RightList);
	break;
      }

      curDesc->Next = RightList;
      curDesc = curDesc->Next;
      RightList = RightList->Next;
    }

    if (curDesc->Node->depth - node->depth > minDist) {

      // Add the number of samples of each size to the
      // FromLeft or FromRight arrays
      if (fromLeft) {
	for (i=0; i < sampleSize-1; i++)
	  FromLeft[i] += curDesc->Node->SamplesRootedHere[i];
      } else {
	for (i=0; i < sampleSize-1; i++)
	  FromRight[i] += curDesc->Node->SamplesRootedHere[i];
      }
    }
  }

  // Calculate the number of samples rooted at this node,
  // of each size
  for (i=1; i < sampleSize; i++) {
    for (j=0; j < i; j++) {
      node->SamplesRootedHere[i] += FromLeft[j] * FromRight[i-j-1];
    }
  }

  // Free the temporary arrays
  free(FromRight);
  free(FromLeft);
#ifdef _MEM_INFO_
  DEC_MEM(2 * (sampleSize-1) * sizeof(double));
#endif

  // The descendents of the children are no longer needed (and weve
  // messed up their lists anyway)
  node->LChild->Descendents = NULL;
  node->RChild->Descendents = NULL;

  // Add the samples rooted here to our running count
  // of valid samples
  curValidSamples += node->SamplesRootedHere[sampleSize-1];

  return curValidSamples;
}

