
#include <stdlib.h>
 
#include "structs.h"
#include "utils.h"
 
extern double minDist, maxDist;
extern unsigned long sampleSize;

#ifdef _MEM_INFO_
extern double CurMemoryUse, MaxMemoryUse;
#endif

// Find the least common ancestor of two nodes.
//
// TODO: Our method here is really dumb and takes up to O(h) time
//       (where h is the height of the tree), in the worst case. We should
//       implement this to run in O(1) time.
TreeNode * Edge_LCA(TreeNode * node1, TreeNode * node2)
{
  ASSERT(node1 && node2, "Edge_LCA was called with a NULL node!");

  for (;;) {
    if (node1 == node2)
      return node1;

    if (node1->depth < node2->depth)
      node2 = node2->Parent;
    else
      node1 = node1->Parent;
  }  
}

// Recursively deletes the subtree rooted here
void Edge_DeleteInducedSubtree(InducedSubtree * node)
{
  if (node) {
    Edge_DeleteInducedSubtree(node->LChild);
    Edge_DeleteInducedSubtree(node->RChild);
    free(node);
#ifdef _MEM_INFO_
    DEC_MEM(sizeof(InducedSubtree));
#endif
  }
}

// Creates a subtree wrapper node for the given tree node
InducedSubtree * Edge_CreateInducedNode(TreeNode * node)
{
  InducedSubtree * newNode;
  newNode = (InducedSubtree *)malloc(sizeof(InducedSubtree));
  ASSERT(newNode, "Out of memory.");
#ifdef _MEM_INFO_
    INC_MEM(sizeof(InducedSubtree));
#endif
  newNode->Node = node;
  newNode->Parent = NULL;
  newNode->LChild = NULL;
  newNode->RChild = NULL;
  return newNode;
}

// This is really easy to do because we know that the samples we
// get are ordered left-to-right, as in the original tree. We can
// just keep adding one node at a time by finding the least common
// ancestor with the right-most node in the induced subtree and
// the next node coming in.
//
// TODO: We could check that the edge length constraints are not
//       violated as the induced subtree is being built instead
//       of traversing the tree afterwards to check this.
InducedSubtree * Edge_BuildInducedSubtree(TreeNodeList * nodes)
{
  InducedSubtree * subtree, * prevNode, * newNode, * lcaSubNode;
  TreeNode * lca;

  ASSERT(nodes && nodes->Next, "Sample contains less than two nodes!");

  // Get the first two nodes off the list and find their LCA
  prevNode = Edge_CreateInducedNode(nodes->Node);
  nodes = nodes->Next;
  newNode = Edge_CreateInducedNode(nodes->Node);
  nodes = nodes->Next;
  subtree = Edge_CreateInducedNode(Edge_LCA(prevNode->Node, newNode->Node));
  subtree->LChild = prevNode;
  subtree->RChild = newNode;
  prevNode->Parent = subtree;
  newNode->Parent = subtree;

  while (nodes) {

    // Pull the next node off the list to be added
    prevNode = newNode;
    newNode = Edge_CreateInducedNode(nodes->Node);
    nodes = nodes->Next;

    // Find the LCA (in the original tree) of this new node 
    // with the right-most node in the induced subtree
    lca = Edge_LCA(prevNode->Node, newNode->Node);

    // Walk up the right branch of the tree until we either get
    // past the root of the induced subtree, or we find the
    // correct spot along the branch to insert the LCA node.
    for (;;) {
      if (!prevNode || prevNode->Node->depth <= lca->depth)
	break;
      prevNode = prevNode->Parent;
    }

    lcaSubNode = Edge_CreateInducedNode(lca);
    if (!prevNode) {

      // New root for the induced subtree
      lcaSubNode->LChild = subtree;
      lcaSubNode->RChild = newNode;
      newNode->Parent = lcaSubNode;
      subtree->Parent = lcaSubNode;
      subtree = lcaSubNode;
    } else {

      // We should never end up with an LCA node thats already
      // in the tree. If we do, something went wrong.
      ASSERT(prevNode->Node != lca, "Error building induced subtree!");

      lcaSubNode->LChild = prevNode->RChild;
      lcaSubNode->RChild = newNode;
      lcaSubNode->Parent = prevNode;
      prevNode->RChild = lcaSubNode;
      lcaSubNode->LChild->Parent = lcaSubNode;
      newNode->Parent = lcaSubNode;      
    }
  }

  return subtree;
}

// Traverses the subtree ensuring that every edge satisfies
// the given constraints.
void Edge_CheckInducedSubtree(InducedSubtree * node)
{
  ASSERT(node, "CheckInducedSubtree() called with a NULL node.");

  if (node->Parent) {
    ASSERT((node->Node->depth - node->Parent->Node->depth) > minDist,
	   "An edge in the sample was too short.");
    ASSERT((node->Node->depth - node->Parent->Node->depth) < maxDist,
	   "An edge in the sample was too long.");
  }

  // Recursively check children
  if (!isLeaf(node->Node)) {
    Edge_CheckInducedSubtree(node->LChild);
    Edge_CheckInducedSubtree(node->RChild);
  }
}

// Prints out the given subtree in Newick format. External calls
// to PrintSubtree should set index to 0.
void Edge_PrintInducedSubtree(InducedSubtree * node, int indent)
{
  int i, k;

  // Add some spaces to show hierarchy
  for (k=0; k < indent; k++)
    printf(" ");

  if (isLeaf(node->Node)) {
    double dist = node->Node->depth - node->Parent->Node->depth;

    printf("\'");
    for (i=0; node->Node->Name[i] != '\0'; i++) {
      char c = node->Node->Name[i];
      if (c == '[' || c == '(')
	c = '<';
      else if (c == ']' || c == ')')
	c = '>';
      else if (c == ';' || c == ':')
	continue;

      printf("%c", c);
    }
	
    printf("\'");
    if (dist < 0.0001)
      printf(":%e", dist);
    else
      printf(":%f", dist);
  } else {

    // Recursively print out child subtrees
    printf("(\n");
    Edge_PrintInducedSubtree(node->LChild, indent+1);
    printf(",\n");
    Edge_PrintInducedSubtree(node->RChild, indent+1);
    printf("\n");

    // Indent accordingly
    for (k=0; k < indent; k++)
      printf(" ");

    if (!node->Parent) {
      printf(");\n");
    } else {
      double dist = node->Node->depth - node->Parent->Node->depth;
      if (dist < 0.0001)
	printf("):%e", dist);
      else
	printf("):%f", dist);
    }
  }
}
