
#include <stdlib.h>
#include <math.h>

#include "structs.h"
#include "utils.h"
#include "mersenne.h"

extern TreeNode * rootNode;
extern double minDist, maxDist;
extern double numValidSamples;
extern double CurMemoryUse, MaxMemoryUse;
extern unsigned long sampleSize;

#ifdef _MEM_INFO_
extern double CurMemoryUse, MaxMemoryUse;
#endif

// Find the valid node descendent of the given node corresponding
// to the given "index". The "index" value is actually an index into
// the list of all valid samples of the given size, so to find the
// right node we increment the index value by the number of valid
// samples of the given size at each node we visit. If the node is not 
// a descendent of this node, the index value is updated, and the function 
// returns NULL.
TreeNode * Edge_FindNode(TreeNode * node, double minDepth, double maxDepth,
			 int samSize, double * index)
{
  TreeNode * result;

  // Check that the node is not too deep
  if (node->depth >= maxDepth)
    return NULL;

  // If its a valid node, update the index value
  if (node->depth > minDepth) {
    *index = (*index) - node->SamplesRootedHere[samSize-1];

    // If this is the desired node, return it
    if ((*index) <= 0.0)
      return node;
  }

  if (isLeaf(node))
    return NULL;

  // If this is an internal node, try the children
  result = Edge_FindNode(node->LChild, minDepth, maxDepth, samSize, index);
  if (result)
    return result;
  result = Edge_FindNode(node->RChild, minDepth, maxDepth, samSize, index);
  if (result)
    return result;

  return NULL;
}

// Does an inorder traversal of the tree, finding the correct sample
// based on the point at which the sum of the samples adds up to
// the given rootIndex (which is between 0 and numValidSamples).
//
// If the root is found, "samRoot" will point to it and -1.0 is
// returned. If not, it will return the current index, after
// subtracting the number of valid samples rooted at each of
// the nodes that have been visited.
double Edge_FindSampleRoot(TreeNode * node, double rootIndex, 
			   TreeNode ** samRoot)
{
  double curIndex;

  // Root must be an internal node
  if (isLeaf(node))
    return rootIndex;

  // Try left subtree
  curIndex = Edge_FindSampleRoot(node->LChild, rootIndex, samRoot);
  if (curIndex == -1.0)
    return -1.0;
  
  // Try this node
  if (curIndex < node->SamplesRootedHere[sampleSize-1]) {
    *samRoot = node;
    return -1.0;
  }
  curIndex -= node->SamplesRootedHere[sampleSize-1];

  // Try right subtree
  return Edge_FindSampleRoot(node->RChild, curIndex, samRoot);
}

// Traverses a given subtree counting up the number of samples
// of size 1, ..., maxSize for all the nodes whose depth is in
// the given range.
void Edge_CountSamples(TreeNode * node, double minDepth, double maxDepth,
		       double * NumSamples, int maxSize)
{
  int i;

  // If we are too far away, all of our descendents will be as well
  if (node->depth >= maxDepth)
    return;

  // Leaves contain exactly one sample of size one
  if (isLeaf(node)) {
    if (node->depth > minDepth)
      NumSamples[0] += 1.0;
    return;
  }

  // Count samples in descendent leaves
  Edge_CountSamples(node->LChild, minDepth, maxDepth, NumSamples, maxSize);
  Edge_CountSamples(node->RChild, minDepth, maxDepth, NumSamples, maxSize);

  // Add samples rooted at this node
  ASSERT(node->SamplesRootedHere, "Internal node has an empty SamplesRootedHere array.");
  if (node->depth >= minDepth) {
    for (i=1; i < maxSize; i++)
      NumSamples[i] = NumSamples[i] + node->SamplesRootedHere[i];
  }
}

// Selects a valid sample of the given size rooted at the
// given node. Returns the leaves of the sample as a linked-list.
TreeNodeList * Edge_SelectSampleOfSize(TreeNode * node, int samSize)
{
  TreeNodeList * Sample, * curNode;
  double * LSamples, * RSamples, * FromLeft;
  double sampleIndex, index;
  TreeNode * LNode, * RNode;
  unsigned long i, j, l;
  double k=0.0;
  double sampleCount = 0.0;

  ASSERT(node, "SelectSampleOfSize() called with a NULL node.");

  if (isLeaf(node)) {
    ASSERT(samSize == 1, "Asking for a sample of size > 1 from a leaf.");

    Sample = (TreeNodeList *)malloc(sizeof(TreeNodeList));
    ASSERT(Sample, "Out of memory.");
#ifdef _MEM_INFO_
    INC_MEM(sizeof(TreeNodeList));
#endif
    Sample->Node = node;
    Sample->Next = NULL;
    return Sample;
  }

  ASSERT(samSize > 1, "Asking for only one leaf from an internal node.");

  // Get the number of samples in the left and right subtrees
  // so we can decide how many leaves to take from each
  LSamples = (double *)malloc((samSize-1) * sizeof(double));
  ASSERT(LSamples, "Out of memory.");
  RSamples = (double *)malloc((samSize-1) * sizeof(double));
  ASSERT(RSamples, "Out of memory.");
#ifdef _MEM_INFO_
  INC_MEM(2*(samSize-1)*sizeof(double));
#endif

  // Initialize all the entries  
  for (i=0; i < samSize-1; i++) {
    LSamples[i] = 0.0;
    RSamples[i] = 0.0;
  }

  // Ask the left and right child how many samples they have
  Edge_CountSamples(node->LChild, node->depth + minDist, node->depth + maxDist,
		    LSamples, samSize-1);
  Edge_CountSamples(node->RChild, node->depth + minDist, node->depth + maxDist,
		    RSamples, samSize-1);

  // Count how many samples there are with each number taken
  // from the left subtree (and the rest from the right)
  FromLeft = (double *)malloc((samSize-1) * sizeof(double));
  ASSERT(FromLeft, "Out of memory.");
#ifdef _MEM_INFO_
  INC_MEM((samSize-1)*sizeof(double));
#endif
  for (i=0; i < samSize-1; i++) {
    FromLeft[i] = LSamples[i] * RSamples[samSize-2-i];
#ifdef _DEBUG_
    sampleCount += FromLeft[i];
#endif
  }

#ifdef _DEBUG_
  ASSERT(sampleCount == node->SamplesRootedHere[samSize-1], "Sample count does not match count from enumeration.");
#endif

  // Pick a sample index
  sampleIndex = node->SamplesRootedHere[samSize-1] * Edge_genrand();
  if (sampleIndex == node->SamplesRootedHere[samSize-1])
    sampleIndex--;

  // Find out how many nodes should be taken from each subtree
  for (i=0;; i++) {
    if (FromLeft[i] >= sampleIndex)
      break;
    sampleIndex -= FromLeft[i];
  }

  ASSERT(i < samSize-1, "Tried to take all the leaves from the right subtree of an internal node.");

  // Find the next node that will be taken from the left and right subtrees
  index = Edge_genrand() * LSamples[i];
  LNode = Edge_FindNode(node->LChild, node->depth + minDist, 
			node->depth + maxDist, i+1, &index);
  ASSERT(LNode, "Failed to find a node in the induced subtree.");

  index = Edge_genrand() * RSamples[samSize-2-i];
  RNode = Edge_FindNode(node->RChild, node->depth + minDist, 
			node->depth + maxDist, samSize-1-i, &index);
  ASSERT(RNode, "Failed to find a node in the induced subtree.");

  // Get the samples from each child
  Sample = Edge_SelectSampleOfSize(LNode, i+1);
  ASSERT(Sample, "Failed to select a subsample.");

  curNode = Sample;
  while (curNode->Next != NULL)
    curNode = curNode->Next;
  curNode->Next = Edge_SelectSampleOfSize(RNode, samSize-1-i);
  ASSERT(curNode->Next, "Failed to select a subsample.");

  // Free the temporary arrays
  free(LSamples);
  free(RSamples);
  free(FromLeft);
#ifdef _MEM_INFO_
  DEC_MEM(3*(samSize-1)*sizeof(double));
#endif

  return Sample;
}

// Picks a root for the sample and then calls SelectSampleOfSize()
// to do all the sampling work.
TreeNodeList * Edge_SelectSample()
{
  TreeNode * sampleRoot = NULL;
  double rootIndex;
  TreeNodeList * Sample, * temp;

  //
  // Pick a root for the sample uniformly at random
  //
  rootIndex = numValidSamples * Edge_genrand();
  if (rootIndex == numValidSamples)
    rootIndex--;

  Edge_FindSampleRoot(rootNode, rootIndex, &sampleRoot);
  ASSERT(sampleRoot, "Failed to find a root for the sample.");

  // 
  // Pick leaves for the sample
  //
  Sample = Edge_SelectSampleOfSize(sampleRoot, sampleSize);
  ASSERT(Sample, "Failed to generate the required sample.");

  return Sample;
}
