
#include <stdlib.h>
#include <stdio.h>
#include "structs.h"
#include "utils.h"

extern unsigned long numLeaves, sampleSize;

#ifdef _MEM_INFO_
extern double CurMemoryUse, MaxMemoryUse;
#endif

// Returns m! (really only useful if m is an integer) 
double Edge_ffact(double m)
{
  double temp = 1.0;
  for (;m > 1.0;m--)
    temp = temp * m;
  return temp;
}

// Returns n choose m (really only useful if both are integers) 
//
// TODO: We might want to precompute all the choose values we
//       need beforehand and then look them up in a table
//       since every calculation has a number between "sampleSize"
//       and "numLeaves" on top and between 2 and "sampleSize" 
//       on the bottom. These could easily be precomputed.
double Edge_fchoose(double n, double m)
{
  if (m == 0 || m > n)
    return 0;
  else if (m == n)
    return 1;

  if (m > n/2.0) {
    return Edge_fchoose(n, n-m);
  } else {
    double i, temp = n / Edge_ffact(m);
    for (i=n-1.0; i > n-m; i--) {
      temp = temp * i;
    }
    return temp;
  }
}

// recursively deletes the list
void Edge_DeleteList(TreeNodeList * node)
{
  if (node) {
    Edge_DeleteList(node->Next);
    free(node);
#ifdef _MEM_INFO_
    DEC_MEM(sizeof(TreeNodeList));
#endif
  }
}

// recursively deletes all storage for the tree
void Edge_DeleteTree(TreeNode * node)
{
  ASSERT(node, "DeleteTree() called with a NULL node.");

  if (node->LChild)
    Edge_DeleteTree(node->LChild);
  if (node->RChild)
    Edge_DeleteTree(node->RChild);
  if (node->Name)
    free(node->Name);
  if (node->SamplesRootedHere)
    free(node->SamplesRootedHere);
  if (node->Descendents)
    Edge_DeleteList(node->Descendents);  
}


#define STATE_NONE                    0
#define STATE_READING_NAME_IN_QUOTES  1
#define STATE_READING_DIST            2
#define MAX_NAME_SIZE                 1024
#define SHORT_NAME_LENGTH             20

// Reads through the input file and constructs a tree,
// returning a pointer to the root of the tree
//
// TODO: This should really be rewritten. Its big and ugly.
//       Should use recursion for subtrees to avoid having
//       to use the "Parent" field for stepping up the tree.
//       Either that or just keep its own stack of parent nodes.
TreeNode * Edge_ReadInTree(FILE * fp, int useShortNames)
{
  TreeNode * curNode = NULL;
  TreeNode * rootNode = NULL;
  char * curName;
  char * curDist;
  int curNameIndex, curDistIndex;
  int wentUp = 0;
  int InternalIndex = 0;
  int curChar = 0, curLine = 0;
  int readingName = 0;
  char c, lastChar;
  int state = STATE_NONE;

  ASSERT(fp, "ReadInTree() called with a NULL file.");                          

  curNode = NULL;
  curName = malloc((MAX_NAME_SIZE+1) * sizeof(char));
  ASSERT(curName, "Out of memory.");
  curDist = malloc(255);
  ASSERT(curName, "Out of memory.");
  curNameIndex = 0;
  curDistIndex = 0;

  numLeaves = 0;

  for (;;) {
    lastChar = c;
    c = fgetc(fp);
    if (feof(fp))
      break;

    if (c == '\n') {
      curChar = 0;
      curLine++;
      continue;
    } else {
      curChar++;
    }

    // File format is:
    //
    // T -> (T:dist,T:dist);
    // T -> 'name'
    //
    // Whitespace of any kind is ignored unless it
    // is part of a node name.

    if (c == '\'') {
      if (state == STATE_READING_NAME_IN_QUOTES) {
	state = STATE_NONE;
      } else {

	state = STATE_READING_NAME_IN_QUOTES;

	// We'll just strip out single quotes. They're too
	// annoying to deal with.
      }
    } else if (state == STATE_READING_NAME_IN_QUOTES) {

      // Entering a name
      if ((!useShortNames && curNameIndex < MAX_NAME_SIZE-1)
          || curNameIndex < SHORT_NAME_LENGTH) {
	curName[curNameIndex] = c;
	curNameIndex++;
      }
    } else if (isspace(c)) {
      continue;
    } else if (c == '(') {

      // New (internal) node. Create one and make it
      // the current node.
      TreeNode * newNode = malloc(sizeof(TreeNode));
      ASSERT(newNode, "Out of memory.");
#ifdef _MEM_INFO_
      INC_MEM(sizeof(TreeNode));
#endif
      state = STATE_NONE;

#ifdef _DEBUG_
      // We dont really need to name the internal nodes, so only
      // do it if we are debugging.
      newNode->Name = (char *)malloc(MAX_NAME_SIZE * sizeof(char));
      ASSERT(newNode->Name, "Out of memory.");
#ifdef _MEM_INFO_
      INC_MEM(MAX_NAME_SIZE * sizeof(char));
#endif
      sprintf(newNode->Name, "Internal%d", InternalIndex);
      InternalIndex++;
//++pac  #elsif
#else
      newNode->Name = NULL;
#endif
      newNode->LChild = NULL;
      newNode->RChild = NULL;
      newNode->Parent = curNode;
      newNode->Descendents = NULL;
      newNode->depth = 0;
      newNode->SamplesRootedHere = NULL;
      if (!rootNode) {
	rootNode = newNode;
      } else if (!curNode->LChild) {
	curNode->LChild = newNode;
      } else {
	curNode->RChild = newNode;
      }
      curNode = newNode;
    } else if (c == ':') {
      curDistIndex = 0;
      state = STATE_READING_DIST;

      if (wentUp == 0) {

	// Finished entering a name. Create a node for
	// this element, and prepare to read distance
	// to parent.
	TreeNode * newNode = malloc(sizeof(TreeNode));
        ASSERT(newNode, "Out of memory.");
#ifdef _MEM_INFO_
        INC_MEM(sizeof(TreeNode));
#endif

	curName[curNameIndex] = '\0';
        newNode->Name = (char *)strdup(curName);
        ASSERT(newNode->Name, "Out of memory.");
#ifdef _MEM_INFO_
        INC_MEM((strlen(curName)+1) * sizeof(char));
#endif
	curNameIndex = 0;      
	newNode->LChild = NULL;
	newNode->RChild = NULL;
	newNode->Parent = curNode;
	newNode->Descendents = NULL;
	newNode->SamplesRootedHere = NULL;
	newNode->depth = 0;
	if (curNode->LChild == NULL) {
	  curNode->LChild = newNode;
	} else {
	  curNode->RChild = newNode;
	}

	numLeaves++;
      }

      wentUp = 0;

      
    } else if (c == ',') {
      state = STATE_NONE;

      curDist[curDistIndex] = '\0';

      // For now, we ll set the depth of a node as being the distance
      // from its parent. This will get fixed when CountAllSamples()
      // is called.
      curNode->LChild->depth = atof(curDist);
      curDistIndex = 0;
    } else if (c == ')') {
      state = STATE_NONE;

      curDist[curDistIndex] = '\0';

      // For now, we ll set the depth of a node as being the distance
      // from its parent. This will get fixed when CountAllSamples()
      // is called.
      curNode->RChild->depth = atof(curDist);
      curDistIndex = 0;

      curNode = curNode->Parent;
      wentUp = 1;
    } else if (c == ';') {
      break;
    } else if (state != STATE_READING_DIST && isprint(c) && c != '[' && c != ']') {
      if ((!useShortNames && curNameIndex < MAX_NAME_SIZE-1)
          || curNameIndex < SHORT_NAME_LENGTH) {
	curName[curNameIndex] = c;
	curNameIndex++;
      }
    } else if (state == STATE_READING_DIST &&
	       (isdigit(c) || c == '.' || c == 'e' || c == '-')) {
      // Entering a distance
      curDist[curDistIndex] = c;
      curDistIndex++;
    } else {
      printf("Error in input file, line %d, char %d. Exiting...\n", curLine, curChar);
      exit(-1);
    }
  }

  free(curName);
  free(curDist);

  return rootNode;
}

// Prints out each node and some info in this format:
//
// <Name>: <DistanceFromParent> (<SamplesRootedHere>) <LeafDescendentsList...>
//
// Uses indentation to show the parent-child relationship
// as we go down the tree
void Edge_PrintTree(TreeNode * node, int indent) 
{
  int i;

  if (!node)
    return;

  ASSERT(node->Name, "Null Name!");

  for (i=0; i < indent; i++)
    printf("  ");

  printf("%s: %.3f ", node->Name, node->depth);
  for (i=0; i < sampleSize; i++)
    printf("%.0f ", node->SamplesRootedHere[i]);

  printf("\n");
  Edge_PrintTree(node->LChild, indent+1);
  Edge_PrintTree(node->RChild, indent+1);
}

