
#include <stdlib.h>

#include "utils.h"
#include "structs.h"

extern unsigned long sampleSize;
extern double minDist, maxDist;
extern TreeNode ** LeafArray;
extern double CurMemoryUse, MaxMemoryUse;
extern unsigned long numLeaves;

// This function takes care of the enumeration of all samples at
// the given node where every leaf came from the same subtree
// (either all in the left subtree, or all in the right)
void Pairwise_DoSamplesFromOneSubtree(TreeNode * node)
{
  CloseNode * LeftClose, * RightClose, * Close;
  FarNode * childFar, * Far;
  SizeNode * childSize, * Size;
  LeftNode * childLeft;
  int fromLeft;
  double ValidSamples;

  ASSERT(node, "DoSamplesFromOneSubtree() called with a NULL node.");

  // We ll need to look through the matrix entries for the two children
  LeftClose = node->LChild->Close;
  RightClose = node->RChild->Close;

  node->Close = NULL;
  for (;LeftClose || RightClose;) {

    // Consider each node in increasing depth order
    if (!LeftClose) {
      fromLeft = 0;
    } else if (!RightClose) {
      fromLeft = 1;
    } else if (LeftClose->ID < RightClose->ID) {
      fromLeft = 1;
    } else {
      fromLeft = 0;
    }

    // Create a new entry in the matrix for this close node
    if (!node->Close) {
      node->Close = (CloseNode *)malloc(sizeof(CloseNode));
      ASSERT(node->Close, "Out of memory.");
      INC_MEM(sizeof(CloseNode));
      Close = node->Close;
    } else {
      Close->Next = (CloseNode *)malloc(sizeof(CloseNode));
      ASSERT(Close->Next, "Out of memory.");
      INC_MEM(sizeof(CloseNode));
      Close = Close->Next;
    }

    // Fill in the entry
    if (fromLeft) {
      Close->ID = LeftClose->ID;
      childFar = LeftClose->Far;
      LeftClose = LeftClose->Next;
    } else {
      Close->ID = RightClose->ID;
      childFar = RightClose->Far;
      RightClose = RightClose->Next;
    }

    Close->Next = NULL;
    Close->Far = NULL;

    // Copy over all the far node entries for this close node
    for (;childFar; childFar = childFar->Next) {

      // Create the entry
      if (!Close->Far) {
	Close->Far = (FarNode *)malloc(sizeof(FarNode));
	ASSERT(Close->Far, "Out of memory.");
	INC_MEM(sizeof(FarNode));
	Far = Close->Far;
      } else {
	Far->Next = (FarNode *)malloc(sizeof(FarNode));
	ASSERT(Far->Next, "Out of memory.");
	INC_MEM(sizeof(FarNode));
	Far = Far->Next;
      }

      Far->ID = childFar->ID;
      Far->Size = NULL;
      Far->Next = NULL;

      // Repeat for each sample size
      childSize = childFar->Size;
      for (; childSize; childSize = childSize->Next) {

	// Create an entry for this sample size
	if (!Far->Size) {
	  Far->Size = (SizeNode *)malloc(sizeof(SizeNode));
	  ASSERT(Far->Size, "Out of memory.");
	  INC_MEM(sizeof(SizeNode));
	  Size = Far->Size;
	} else {
	  Size->Next = (SizeNode *)malloc(sizeof(SizeNode));
	  ASSERT(Size->Next, "Out of memory.");
	  INC_MEM(sizeof(SizeNode));
	  Size = Size->Next;
	}
	Size->Next = NULL;
	Size->SamSize = childSize->SamSize;

	// Count all the valid samples of this size from
	// the child node (with the current close/far nodes)
	ValidSamples = 0.0;
	for (childLeft=childSize->Left; childLeft; childLeft=childLeft->Next)
	  ValidSamples += childLeft->Samples;

	ASSERT(ValidSamples > 0.0, "Sample list with no valid samples.");

	// Either all the entries are from our left subtree, or all
	// are from the right subtree, so insert this entry here
	Size->Left = (LeftNode *)malloc(sizeof(LeftNode));
	ASSERT(Size->Left, "Out of memory.");
	INC_MEM(sizeof(LeftNode));
	Size->Left->Samples = ValidSamples;
	Size->Left->Next = NULL;
	if (fromLeft)
	  Size->Left->FromLeft = Size->SamSize;
	else
	  Size->Left->FromLeft = 0;
      }
    }
  }
}

// Simply takes the descendent lists from the left and right child
// and merges them into the descendent list for the given node.
// Maintains the increasing depth order of the leaves.
void Pairwise_MergeDescendents(TreeNode * node)
{
  TreeNodeList *LeftList, *RightList, *curDesc, *temp;
  int fromLeft;

  ASSERT(node, "MergeDescendents() called with a NULL node.");

  // Start at the beginning of both lists
  LeftList = node->LChild->Descendents;
  RightList = node->RChild->Descendents;

  node->Descendents = NULL;
  for (;;) {

    // If both lists are exhausted, we are done
    if (!LeftList && !RightList) {
      curDesc->Next = NULL;
      break;
    }

    // Grab the shallowest leaf still in the lists
    if (!LeftList) {
      fromLeft = 0;
    } else if (!RightList) {
      fromLeft = 1;
    } else if (ID(LeftList) <= ID(RightList)) {
      fromLeft = 1;
    } else {
      fromLeft = 0;
    }

    // Add this leaf to the descendent list
    if (fromLeft) {
      if (!node->Descendents) {
	node->Descendents = LeftList;
	curDesc = node->Descendents;
      } else {
	curDesc->Next = LeftList;
	curDesc = curDesc->Next;
      }

      LeftList = LeftList->Next;
    } else {
      if (!node->Descendents) {
	node->Descendents = RightList;
	curDesc = node->Descendents;
      } else {
	curDesc->Next = RightList;
	curDesc = curDesc->Next;
      }

      RightList = RightList->Next;
    }
  }
}

// Looks through the Close list for this node to find the node with
// the correct closeID (this node must be in the list - 
// DoSamplesFromOneSubtree guarantees this). Then finds the correct
// far node (with the given farID). Tries to start looking with the
// given close/far nodes to be more efficient.
void Pairwise_FindOrCreateCloseFarNodes(TreeNode * node, 
			       CloseNode ** Close, FarNode ** Far,
			       unsigned long closeID, unsigned long farID)
{
  ASSERT(node && Close && *Close && closeID <= numLeaves && farID <= numLeaves,
	 "Invalid parameter to FindOrCreateCloseFarNodes().");

  if ((*Close)->ID != closeID) {
    
    // If we are too far in the list, start back at the
    // beginning. This should never happen (???)
    if ((*Close)->ID > closeID)
      *Close = node->Close;
    
    // Find the close node
    for (; (*Close)->ID != closeID; *Close = (*Close)->Next)
      ;

    ASSERT(*Close, "Could not find the required close node.");
    
    *Far = (*Close)->Far;
  }

  // If we are given a NULL far node, start at the beginning of the list
  if (!(*Far))
    *Far = (*Close)->Far;

  // Check to see if we are at the correct node
  if ((*Far)->ID != farID) {
    
    // If we need to start back at the beginning, do it now	    
    if (farID < (*Far)->ID)
      *Far = (*Close)->Far;

    if (farID < (*Far)->ID) {
      
      // This node should be added to the front of the list
      *Far = (FarNode *)malloc(sizeof(FarNode));
      ASSERT(*Far, "Out of memory.");
      INC_MEM(sizeof(FarNode));
      (*Far)->Next = (*Close)->Far;
      (*Far)->ID = farID;
      (*Far)->Size = NULL;
      (*Close)->Far = *Far;
    } else {
      
      for (;; *Far = (*Far)->Next) {
	
	ASSERT(*Far, "Unexpected error.");
	
	// Found the node
	if ((*Far)->ID == farID)
	  break;
		
	// Add this node onto the end of the list
	if (!(*Far)->Next) {
	  (*Far)->Next = (FarNode *)malloc(sizeof(FarNode));
	  ASSERT((*Far)->Next, "Out of memory.");
	  INC_MEM(sizeof(FarNode));
	  *Far = (*Far)->Next;
	  (*Far)->ID = farID;
	  (*Far)->Size = NULL;
	  (*Far)->Next = NULL;
	  break;
	}

	// Found the spot to insert this node
	if (farID < (*Far)->Next->ID) {
	  FarNode * newFar = (FarNode *)malloc(sizeof(FarNode));
	  ASSERT(newFar, "Out of memory.");
	  INC_MEM(sizeof(FarNode));
	  newFar->Next = (*Far)->Next;
	  (*Far)->Next = newFar;
	  *Far = newFar;
	  (*Far)->ID = farID;
	  (*Far)->Size = NULL;
	  break;
	}
      }
    }
  }
}

// Looks through the sample size list for the given far node, trying
// to find the node with the given sample size. If the node does
// not exist, adds in a new node maintaining the sorted order of
// the list. Tries to start looking at the given sample size node
// to be more efficient.
//
// Returns 1 if the size node is changed by the procedure. If the
// given node was the correct one, returns 0.
int Pairwise_FindOrCreateSizeNode(FarNode * Far, SizeNode ** Size,
				  unsigned long SamSize)
{
  int changed = 0;

  ASSERT(Far && Size && SamSize > 0 && SamSize <= sampleSize,
	 "Invalid parameter to FindOrCreateSizeNode().");

  // If there is no sample size list for the given far node, make
  // this the first node in the list
  if (!Far->Size) {
      *Size = (SizeNode *)malloc(sizeof(SizeNode));
      ASSERT(*Size, "Out of memory.");
      INC_MEM(sizeof(SizeNode));
      (*Size)->Next = NULL;
      (*Size)->SamSize = SamSize;
      (*Size)->Left = NULL;
      Far->Size = *Size;    
      changed = 1;
  } else {

    // If we are given a NULL sample size node, start at the
    // beginning of the list
    if (*Size == NULL) {
      *Size = Far->Size;
      changed = 1;
    }

    // Check to see if we are at the correct node
    if ((*Size)->SamSize != SamSize) {
      changed = 1;

      // If we need to start back at the beginning of the list
      // (we are too far along), do it now
      if (SamSize < (*Size)->SamSize)
	*Size = Far->Size;

      // Add the entry to the front of the list, if this is
      // where it belongs
      if (SamSize < (*Size)->SamSize) {
	*Size = (SizeNode *)malloc(sizeof(SizeNode));
	ASSERT(*Size, "Out of memory.");
	INC_MEM(sizeof(SizeNode));
	(*Size)->Next = Far->Size;
	(*Size)->SamSize = SamSize;
	(*Size)->Left = NULL;
	Far->Size = *Size;
      } else {
	
	for (;; *Size = (*Size)->Next) {
	  
	  ASSERT(*Size, "Unexpected error.");

	  // Found the correct node
	  if ((*Size)->SamSize == SamSize)
	    break;

	  // Got to the end of the list, add the node on to the end
	  if (!(*Size)->Next) {
	    (*Size)->Next = (SizeNode *)malloc(sizeof(SizeNode));
	    ASSERT((*Size)->Next, "Out of memory.");
	    INC_MEM(sizeof(SizeNode));
	    *Size = (*Size)->Next;
	    (*Size)->SamSize = SamSize;
	    (*Size)->Next = NULL;
	    (*Size)->Left = NULL;
	    break;
	  }

	  // Found the spot to insert this node
	  if (SamSize < (*Size)->Next->SamSize) {
	    SizeNode * newSize = (SizeNode *)malloc(sizeof(SizeNode));
	    ASSERT(newSize, "Out of memory.");
	    INC_MEM(sizeof(SizeNode));
	    newSize->Next = (*Size)->Next;
	    (*Size)->Next = newSize;
	    *Size = newSize;
	    (*Size)->SamSize = SamSize;
	    (*Size)->Left = NULL;
	    break;
	  }
	}
      }
    }
  }

  return changed;
}

// Looks through the left list for the given size node, trying to
// find the node with the given number of leaves taken from the
// left subtree. If the node does not exist, adds in a new node
// maintaining the sorted order of the list. Tries to start looking
// at the given left node to be more efficient.
void Pairwise_FindOrCreateLeftNode(SizeNode * Size, LeftNode ** Left,
				   unsigned long FromLeft)
{
  ASSERT(Size && Left && FromLeft <= Size->SamSize,
	 "Invalid parameter to FindOrCreateLeftNode().");

  // If there is no left list for the given node, make this
  // the first entry in the list
  if (!Size->Left) {
    *Left = (LeftNode *)malloc(sizeof(LeftNode));
    ASSERT(*Left, "Out of memory.");
    INC_MEM(sizeof(LeftNode));
    (*Left)->Next = NULL;
    (*Left)->FromLeft = FromLeft;
    (*Left)->Samples = 0.0;
    Size->Left = *Left;
  } else{

    // If we are given a NULL left node, start a the beginning
    // of the list
    if (*Left == NULL)
      *Left = Size->Left;

    // Check to see if this is the correct node
    if ((*Left)->FromLeft != FromLeft) {

      // If we need to start at the beginning of the list (we are
      // too far along), do it now
      if (FromLeft < (*Left)->FromLeft)
	*Left = Size->Left;
      
      // If this entry should go at the front of the list, insert it
      if (FromLeft < (*Left)->FromLeft) {
	*Left = (LeftNode *)malloc(sizeof(LeftNode));
	ASSERT(*Left, "Out of memory.");
	INC_MEM(sizeof(LeftNode));
	(*Left)->Next = Size->Left;
	(*Left)->FromLeft = FromLeft;
	(*Left)->Samples = 0.0;
	Size->Left = *Left;
      } else {
	
	for (;; *Left = (*Left)->Next) {

	  ASSERT(*Left, "Unexpected error.");

	  // Found the correct node
	  if ((*Left)->FromLeft == FromLeft)
	    break;
	  
	  // Got to the end of the list and never found the node.
	  // Add it onto the end of the list.
	  if (!(*Left)->Next) {
	    (*Left)->Next = (LeftNode *)malloc(sizeof(LeftNode));
	    ASSERT((*Left)->Next, "Out of memory.");
	    INC_MEM(sizeof(LeftNode));
	    *Left = (*Left)->Next;
	    (*Left)->FromLeft = FromLeft;
	    (*Left)->Next = NULL;
	    (*Left)->Samples = 0.0;
	    break;
	  }
	  
	  // Found the spot to insert this node
	  if (FromLeft < (*Left)->Next->FromLeft) {
	    LeftNode * newLeft = (LeftNode *)malloc(sizeof(LeftNode));
	    ASSERT(newLeft, "Out of memory.");
	    INC_MEM(sizeof(LeftNode));
	    newLeft->Next = (*Left)->Next;
	    (*Left)->Next = newLeft;
	    *Left = newLeft;
	    (*Left)->FromLeft = FromLeft;
	    (*Left)->Samples = 0.0;
	    break;
	  }
	}
      }
    }
  }
}

// Traverses the tree filling in the matrix of valid samples for
// each node in the tree. External calls to this function should
// set isRoot to be 1, so that the function knows to delete the
// descendent lists instead of trying to merge them to be passed
// up the tree.
void Pairwise_CountSamples(TreeNode * node, int isRoot)
{
  unsigned long closeID, farID, SamSize;
  CloseNode *p_l, *p_r, *Close;
  FarNode *q_l, *q_r, *Far;
  SizeNode *Size, *lSize, *rSize;
  LeftNode *Left, *lLeft, *rLeft;
  double LeftSam, RightSam;
  int foundNodes;

  ASSERT(node, "CountSamples() called with a NULL node.");

  // Leaves already have their enumeration taken care of so we
  // just need to put this node in its own descendent list so that
  // it can be merged up the tree.
  if (isLeaf(node)) {
    node->Descendents = (TreeNodeList *)malloc(sizeof(TreeNodeList));
    ASSERT(node->Descendents, "Out of memory.");
    INC_MEM(sizeof(TreeNodeList));
    node->Descendents->Node = node;
    node->Descendents->Next = NULL;
    return;
  }

  // Recursively enumerate lower nodes in the tree first
  Pairwise_CountSamples(node->LChild, 0);
  Pairwise_CountSamples(node->RChild, 0);

  // First handle all valid samples where all the leaves are
  // in the same subtree
  Pairwise_DoSamplesFromOneSubtree(node);

  Close = node->Close;
  ASSERT(Close, "DoSamplesFromOneSubtree() failed.");
  Far = NULL;

  //
  // Now, we ll try combining samples from the left and right subtrees
  //

  // Pick a close/far pair from both subtrees
  for (p_l = node->LChild->Close; p_l; p_l = p_l->Next) {
    for (q_l = p_l->Far; q_l; q_l = q_l->Next) {
      for (p_r = node->RChild->Close; p_r; p_r = p_r->Next) {

	// Check that the close nodes are not too close
	if (L_DEPTH(p_l) + L_DEPTH(p_r) - 2.0*node->depth <= minDist)
	  continue;

	for (q_r = p_r->Far; q_r; q_r = q_r->Next) {

	  // Check that the far nodes are not too far
          if (L_DEPTH(q_l) + L_DEPTH(q_r) - 2.0*node->depth >= maxDist)
	    break;
	  
	  // Find out if the closer/further node came from the left
	  // or right subtree
	  closeID = (L_DEPTH(p_l) <= L_DEPTH(p_r)) ? p_l->ID : p_r->ID;
	  farID = (L_DEPTH(q_l) >= L_DEPTH(q_r)) ? q_l->ID : q_r->ID;

	  //
	  // Try all sample sizes
	  //
	  foundNodes = 0;
	  for (lSize=q_l->Size; lSize; lSize=lSize->Next) {
	    for (rSize=q_r->Size; rSize; rSize=rSize->Next) {

	      // Stop when the sample is too big
	      SamSize = lSize->SamSize + rSize->SamSize;
	      if (SamSize > sampleSize)
		break;

	      // If we haven t found the close/far node yet for this
	      // sample, do it now.
	      //
	      // NOTE: We wait until this point to look for the nodes,
	      //       since we don t want to bother if there are no
	      //       valid samples for this close/far pair.
	      if (!foundNodes) {
		Pairwise_FindOrCreateCloseFarNodes(node, &Close, &Far, closeID, farID);
		ASSERT(Close && Far, "Failed to find or create close and far nodes.");
		Size = NULL;
		Left = NULL;
		foundNodes = 1;
	      }

	      // Find the entry for this sample size
	      if (Pairwise_FindOrCreateSizeNode(Far, &Size, SamSize))
		Left = NULL;
	      ASSERT(Size, "Failed to find or create size node.");

	      // Find the entry for this many nodes from the left subtree
	      Pairwise_FindOrCreateLeftNode(Size, &Left, lSize->SamSize);
	      ASSERT(Left, "Failed to find or create left node.");

	      // Count the number of samples from each side
	      LeftSam = 0.0;
	      for (lLeft=lSize->Left; lLeft; lLeft=lLeft->Next)
		LeftSam += lLeft->Samples;
	      RightSam = 0.0;
	      for (rLeft=rSize->Left; rLeft; rLeft=rLeft->Next)
		RightSam += rLeft->Samples;

	      // Add these samples to the matrix
	      Left->Samples += LeftSam * RightSam;
	    }
	  }
	}
      }
    }
  }
  
  // If we are finished enumerating, free the descendent lists.
  // If not, merge the descendents so that they can be used by
  // our parent node.
  if (isRoot) {
    Pairwise_DeleteList(node->LChild->Descendents);
    Pairwise_DeleteList(node->RChild->Descendents);
    DEC_MEM(numLeaves * sizeof(TreeNodeList));
  } else {
    Pairwise_MergeDescendents(node);
    ASSERT(node->Descendents, "MergeDescendents() failed.");
  }
}

// Just loops through the sample matrix at the root of the
// tree to find the total number of valid samples of the
// given size that are in the tree.
double Pairwise_GetNumValidSamples(TreeNode *node) {
  double samples = 0.0;
  CloseNode *Close;
  FarNode *Far;
  SizeNode *Size;
  LeftNode *Left;

  ASSERT(node, "GetNumValidSamples() called with a NULL node.");

  for (Close=node->Close; Close; Close=Close->Next) {
    for (Far=Close->Far; Far; Far=Far->Next) {
      for (Size=Far->Size; Size; Size=Size->Next) {
	if (Size->SamSize == sampleSize) {
	  for (Left=Size->Left; Left; Left=Left->Next)
	    samples += Left->Samples;
	  break;
	}
      }
    }
  }

  return samples;
}
