
#include <stdlib.h>

#include "utils.h"
#include "structs.h"
#include "mersenne.h"

extern double minDist, maxDist;
extern unsigned long numLeaves, sampleSize;
extern TreeNode ** LeafArray;
extern double CurMemoryUse, MaxMemoryUse;

// (Recursively) deletes all the storage for the given subtree
void Pairwise_DeleteInducedSubtree(InducedSubtree * node)
{
  if (node->LChild)
    Pairwise_DeleteInducedSubtree(node->LChild);
  if (node->RChild)
    Pairwise_DeleteInducedSubtree(node->RChild);
  free(node);
}

// Finds the least common ancestor of the given two nodes in
// the original tree. Runs in O(h) time (where h is the height
// of the tree), which is far from the optimal O(1) time that
// we could have, but good enough for now.
TreeNode * Pairwise_LCA(TreeNode * node1, TreeNode * node2)
{
  for (;;) {

    ASSERT(node1 && node2, "Problems looking for a least common ancestor in the induced subtree.");

    if (node1 == node2)
      return node1;
 
    if (node1->depth < node2->depth)
      node2 = node2->Parent;
    else
      node1 = node1->Parent;
  }
}

// Creates an induced subtree node as a wrapper for the given node
InducedSubtree * Pairwise_CreateInducedNode(TreeNode * node)
{
  InducedSubtree * newNode;
  newNode = (InducedSubtree *)malloc(sizeof(InducedSubtree));
  newNode->Node = node;
  newNode->Parent = NULL;
  newNode->LChild = NULL;
  newNode->RChild = NULL;
  newNode->closeD = newNode->closeN = 0.0;
  newNode->farD = newNode->farN = 0.0;
  return newNode;
}

#define DIST(x,y) (y->Node->depth - x->Node->depth)

// This is really easy to do because we know that the samples we
// get are ordered left-to-right, as in the original tree. We can
// just keep adding one node at a time by finding the least common
// ancestor with the right-most node in the induced subtree and
// the next node coming in.
InducedSubtree * Pairwise_BuildAndCheckInducedSubtree(TreeNodeList * nodes)
{
  InducedSubtree * subtree, * newNode1, * newNode2, * lcaSubNode;
  TreeNode * lca;  

  // Get the first two nodes off the list and find their LCA
  newNode1 = Pairwise_CreateInducedNode(nodes->Node);
  nodes = nodes->Next;
  newNode2 = Pairwise_CreateInducedNode(nodes->Node);
  nodes = nodes->Next;
  subtree = Pairwise_CreateInducedNode(Pairwise_LCA(newNode1->Node, 
						    newNode2->Node));
  subtree->LChild = newNode1;
  subtree->RChild = newNode2;
  if (newNode1->Node->depth < newNode2->Node->depth) {
    subtree->closeD = newNode1->Node->depth;
    subtree->farD = newNode2->Node->depth;
  } else {
    subtree->closeD = newNode2->Node->depth;
    subtree->farD = newNode1->Node->depth;
  }

  ASSERT(DIST(subtree, newNode1) + DIST(subtree, newNode2) > minDist &&
	 DIST(subtree, newNode1) + DIST(subtree, newNode2) < maxDist,
	 "Sample violates the given constraints!");

  newNode1->Parent = subtree;
  newNode2->Parent = subtree;
 
  while (nodes) {
 
    // Pull the next node off the list to be added
    newNode1 = newNode2;
    newNode2 = Pairwise_CreateInducedNode(nodes->Node);
    nodes = nodes->Next;
 
    // Find the LCA of this new node with the right-most
    // node in the induced subtree
    lca = Pairwise_LCA(newNode1->Node, newNode2->Node);
 
    // Walk up the right branch of the tree until we either get
    // past the root of the induced subtree, or we find the
    // correct spot along the branch to insert the LCA node.
    for (;;) {
      if (newNode1 == NULL || newNode1->Node->depth <= lca->depth)
        break;
      
      if (newNode1->Parent) {
	newNode1->Parent->farD = newNode1->farD + 
	  DIST(newNode1->Parent, newNode1);
	newNode1->Parent->closeD = newNode1->closeD +
	  DIST(newNode1->Parent, newNode1);
      }

      newNode1 = newNode1->Parent;
    }
 
    if (newNode1 == NULL) {
 
      // New root for the induced subtree
      lcaSubNode = Pairwise_CreateInducedNode(lca);
      lcaSubNode->LChild = subtree;
      lcaSubNode->RChild = newNode2;
      lcaSubNode->closeD = Pairwise_fmin(subtree->closeD + 
					 DIST(lcaSubNode, subtree),
					 DIST(lcaSubNode, newNode2));
      lcaSubNode->farD = Pairwise_fmax(subtree->farD + 
				       DIST(lcaSubNode, subtree),
				       DIST(lcaSubNode, newNode2));
      newNode2->Parent = lcaSubNode;
      newNode2->closeN = subtree->closeD + DIST(lcaSubNode, subtree) +
	DIST(lcaSubNode, newNode2);
      newNode2->farN = subtree->farD + DIST(lcaSubNode, subtree) +
	DIST(lcaSubNode, newNode2);

      subtree->Parent = lcaSubNode;
      subtree = lcaSubNode;
    } else {

      // The LCA node should be added right below newNode1 
      lcaSubNode = Pairwise_CreateInducedNode(lca);
      lcaSubNode->LChild = newNode1->RChild;
      lcaSubNode->RChild = newNode2;
      lcaSubNode->Parent = newNode1;
      newNode1->RChild = lcaSubNode;
      lcaSubNode->LChild->Parent = lcaSubNode;
      lcaSubNode->RChild->Parent = lcaSubNode;

      //
      // Update the close/far values for the new node, the LCA node,
      // and the parent of the LCA node
      //
      if (newNode1 == subtree) {

	// If newNode1 is the root of the subtree, then it has no
	// non-descendents, so consider only the descendents
	newNode2->closeN = DIST(lcaSubNode, newNode2) +
	  newNode1->closeD - DIST(newNode1, lcaSubNode);
	newNode2->farN = DIST(lcaSubNode, newNode2) +
	  newNode1->farD - DIST(newNode1, lcaSubNode);

      } else {

	// The closest non-descendent of newNode2 will either be
	// a descendent or a non-descendent of newNode1
	newNode2->closeN = DIST(lcaSubNode, newNode2) + 
	  Pairwise_fmin(newNode1->closeD - DIST(newNode1, lcaSubNode),
			newNode1->closeN + DIST(newNode1, lcaSubNode));
	newNode2->farN = DIST(lcaSubNode, newNode2) + 
	  Pairwise_fmax(newNode1->farD - DIST(newNode1, lcaSubNode),
			newNode1->farN + DIST(newNode1, lcaSubNode));
      }

      newNode1->closeD = Pairwise_fmin(newNode1->closeD, 
				       DIST(newNode1, newNode2));
      newNode1->farD = Pairwise_fmax(newNode1->farD, DIST(newNode1, newNode2));
      
      lcaSubNode->closeD = Pairwise_fmin(DIST(lcaSubNode, newNode2),
					 lcaSubNode->LChild->closeD +
					 DIST(lcaSubNode, lcaSubNode->LChild));
      lcaSubNode->farD = Pairwise_fmax(DIST(lcaSubNode, newNode2),
				       lcaSubNode->LChild->farD +
				       DIST(lcaSubNode, lcaSubNode->LChild));
      lcaSubNode->closeN = newNode1->closeN + DIST(newNode1, lcaSubNode);
      lcaSubNode->farN = newNode1->farN + DIST(newNode1, lcaSubNode);

      ASSERT(newNode2->closeN > minDist && newNode2->farN < maxDist,
	     "Sample violates the given constraints!");
	
    }
  }
 
  return subtree;
}

// Print out the given number of spaces
void Pairwise_AddIndent(int indent)
{
  int i;
 
  for (i=0; i < indent; i++)
    printf(" ");
}

// Prints out the given induced subtree in Newick format.
// External calls to PrintSubtree should set indent to 0.
void Pairwise_PrintSubtree(InducedSubtree * node, int indent)
{
  int i;
 
  // Put in a few spaces to show nesting
  Pairwise_AddIndent(indent);
 
  if (isLeaf(node->Node)) {
    double dist = node->Node->depth - node->Parent->Node->depth;
    printf("\'");
    for (i=0; node->Node->Name[i] != '\0'; i++) {
      char c = node->Node->Name[i];
      if (c == '[' || c == '(')
        c = '<';
      else if (c == ']' || c == ')')
        c = '>';
      else if (c == ';' || c == ':')
        continue;
      else if (c == ' ' || c == '\t')
        c = '_';
 
      printf("%c", c);
    }
 
    printf("\'");

    // Print really small numbers in scientific notation
    if (dist < 0.0001)
      printf(":%e", dist);
    else
      printf(":%f", dist);
  } else {
    printf("(\n");
    Pairwise_PrintSubtree(node->LChild, indent+1);
    printf(",\n");
    Pairwise_PrintSubtree(node->RChild, indent+1);
    printf("\n");
    Pairwise_AddIndent(indent);
 
    if (node->Parent == NULL) {
      printf(");\n");
    } else {
      double dist = node->Node->depth - node->Parent->Node->depth;

      // Print really small numbers in scientific notation
      if (dist < 0.0001)
        printf("):%e", dist);
      else
        printf("):%f", dist);
    }
  }
}


