
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <sys/time.h>
#include <sys/resource.h>

#include "structs.h"
#include "utils.h"
#include "mersenne.h"
#include "enum.h"
#include "sample.h"
#include "induced.h"

static TreeNode * rootNode = NULL;
unsigned long numLeaves;
unsigned long sampleSize, numSamples;
double minDist, maxDist;
TreeNode ** LeafArray;
double ** Distances;
double CurMemoryUse, MaxMemoryUse;


// Just prints out how the program should be run and exits.
void Pairwise_ShowUsage(char * command)
{
  printf("Usage: %s (options) <tree_file> (<sample_size> <num_samples> <min_dist> <max_dist>)\n\n", command);
  printf("options include:\n");
  printf("\t-q\tquiet mode, samples will not be displayed\n");
  printf("\t-s\tuse short taxon names (truncate to 20 characters)\n");
  printf("\t-v\tverify that the sample satisfies the constraints\n");
  printf("\t-t\toutput Newick tree instead of sample list\n");
  printf("\t-b\toutput Newick tree and sample list\n");
  printf("\t-i\tdisplay some basic summary info\n");
  exit (-1);
}

// The user can include some options (indicated with a '-'), followed
// by the file name. The can also specify the number of samples, the 
// sample size, the minimum pairwise distance, and the maximum pairwise 
// distance on the command line, although if any of these are omitted
// we pick default values for them. Check out the ShowUsage() function
// above for the exact format and the available options.
int main(int argc, char * argv[])
{
  FILE * fp = NULL;
  double ValidSamples;
  unsigned long i;
  TreeNodeList *Sample, *curNode, *curNode2;
  int quiet = 0;
  int onlyTree = 0;
  int treeAndList = 0;
  int info = 0;
  int verify = 0;
  int shortNames = 0;

  CurMemoryUse = 0.0;
  MaxMemoryUse = 0.0;

  // Set the randomizer seed
  Pairwise_sgenrand(time(0));

  // Set some default values
  numSamples = 1;
  sampleSize = 2;
  minDist = 0.0;
  maxDist = 20.0;

  // Check for options first
  for (i=1; i < argc; i++) {
    if (argv[i][0] == '-') {
      if (argv[i][1] == 'q')
	quiet = 1;
      else if (argv[i][1] == 't')
	onlyTree = 1;
      else if (argv[i][1] == 'b')
	treeAndList = 1;
      else if (argv[i][1] == 'i')
	info = 1;
      else if (argv[i][1] == 'v')
	verify = 1;
      else if (argv[i][1] == 's')
	shortNames = 1;
      else
	Pairwise_ShowUsage(argv[0]);
    } else {
      break;
    }
  }

  // They didnt include a file name
  if (i == argc)
    Pairwise_ShowUsage(argv[0]);

  // Allow the user to specify other options
  if (argc > i+1)
    sampleSize = atoi(argv[i+1]);

  ASSERT(sampleSize >= 2, "You must select a sample size of at least 2.")

  if (argc > i+2)
    numSamples = atoi(argv[i+2]);

  if (argc > i+3)
    minDist = atof(argv[i+3]);
  if (argc >= i+4)
    maxDist = atof(argv[i+4]);

  ASSERT(minDist < maxDist, "You must select a maximum distance that is larger than the minimum distance.");

  // Open up the tree file
  fp = fopen(argv[i], "r");
  ASSERT2(fp, "Could not open file '%s'.", argv[i])

  // This will read in the tree so that the global
  // variable "rootNode" points to the root
  rootNode = Pairwise_ReadInTree(fp, shortNames);
  fclose(fp);
  ASSERT(rootNode, "Unexpected error reading file.");
  
  // Do whatever preprocessing needs to be done on the tree
  Pairwise_PrepareAllNodes(rootNode);

  // Enumerate all the samples
  Pairwise_CountSamples(rootNode, 1);

  // Find out how many valid samples there are total
  ValidSamples = Pairwise_GetNumValidSamples(rootNode);
  ASSERT(ValidSamples > 0.0, "Sorry, but there are no valid samples satisfying the given constraints.");

 // Generate the required number of samples
  for (i=0; i < numSamples; i++) {

    // Pick a sample
    Sample = Pairwise_SelectSample(rootNode, numLeaves, numLeaves, sampleSize);
    ASSERT2(Sample, "Unexpected error generating sample %d.", i+1);

    if (!quiet) {
      printf("Sample %d:\n", i+1);

      // Print out the leaves in the sample
      if (!onlyTree) {
	curNode = Sample;
	while (curNode) {
	  printf("  %s\n", curNode->Node->Name);
	  curNode = curNode->Next;
	}
      }

      if (verify || onlyTree || treeAndList) {

	InducedSubtree * Subtree;

	// Construct the subtree induced by the sample and verify
	// that it satisfies the given constraints;
	Subtree = Pairwise_BuildAndCheckInducedSubtree(Sample);

	// If the user wants the subtree printed out in Newick
	// format, do it now
	if (onlyTree || treeAndList)
	  Pairwise_PrintSubtree(Subtree, 0);

	Pairwise_DeleteInducedSubtree(Subtree);
      } 
    }

    Pairwise_DeleteList(Sample);
    DEC_MEM(sampleSize * sizeof(TreeNodeList));
  }

  // Print some summary data, if the user wants it
  if (info) {
    printf("Number of leaves: %d\n", numLeaves);
    printf("Valid samples: %e (out of %e possible samples)\n", ValidSamples,
	   Pairwise_fchoose((double)numLeaves, (double)sampleSize));
    printf("Memory use: %.0f bytes\n", MaxMemoryUse);
    
   }

  // Free up all the memory we allocated
  free(LeafArray);
  Pairwise_DeleteTree(rootNode);
  
  return 0;
}
