
#include <stdlib.h>

#include "utils.h"
#include "structs.h"
#include "mersenne.h"

extern double minDist, maxDist;
extern unsigned long numLeaves, sampleSize;
extern TreeNode ** LeafArray;
extern double CurMemoryUse, MaxMemoryUse;

// General purpose function for counting and sampling from a given
// node. If the function is called with *samples < 0.0 then we assume
// we are enumerating. In this case we count all samples of size
// "subSize" where no leaf is closer than "closeLeaf" and none is
// further than "farLeaf". Returns the number of valid samples in
// *samples. If instead the function is called with *samples >= 0.0
// then we assume we are trying to pick a sample so we use this
// value as the index and find the correct sample. The new close
// leaf and far leaf IDs are returned as well as the number of
// leaves that should be taken from the left subtree.
void Pairwise_LoopThroughSamples(TreeNode * node, unsigned long closeLeaf,
			unsigned long farLeaf, unsigned long subSize,
			double * samples, unsigned long *NewCloseID,
			unsigned long *NewFarID, unsigned long *FromLeft)
{
  int counting, isCloseDesc, isFarDesc;
  CloseNode *Close;
  FarNode *Far;
  SizeNode *Size;
  LeftNode *Left;

  ASSERT(node && closeLeaf <= numLeaves && farLeaf <= numLeaves &&
	 subSize > 0 && subSize <= sampleSize && samples,
	 "Invalid parameter to LoopThroughSamples().");

  // Find out whether we are counting samples, or trying to select one
  if (*samples < 0.0) { 
    counting = 1;
    *samples = 0.0;
  } else {
    counting = 0;
  }

  // Find out if the close and far leaves are descendents
  // of this node
  isCloseDesc = isFarDesc = 0;
  for (Close = node->Close; Close; Close = Close->Next) {
    if (Close->ID == closeLeaf) {
      isCloseDesc = 1;
      if (isFarDesc)
	break;
    }
    if (Close->ID == farLeaf) {
      isFarDesc = 1;
      if (isCloseDesc)
	break;
    }
  }

  for (Close = node->Close; Close; Close = Close->Next) {

    // If the close leaf is a descendent of this node,
    // ignore all the other leaves
    if (isCloseDesc && Close->ID != closeLeaf)
      continue;

    // Ignore any leaves that are shallower than closeLeaf
    if (closeLeaf != numLeaves && Close->ID < closeLeaf)
      continue;

    // Ignore any leaves that are too close to closeLeaf
    if (!isCloseDesc && closeLeaf != numLeaves &&
	L_DEPTH(Close)+ID_DEPTH(closeLeaf)-2.0*node->Parent->depth <= minDist)
      continue;

    for (Far = Close->Far; Far; Far = Far->Next) {

      // If the far leaf is a descendent of this node, ignore
      // all the other leaves
      if (isFarDesc && Far->ID != farLeaf)
	continue;

      // Ignore any leaves that are deeper than farLeaf
      if (farLeaf != numLeaves && Far->ID > farLeaf)
	break;

      // Ignore any leaves that are too far from farLeaf
      if (!isFarDesc && farLeaf != numLeaves &&
	  L_DEPTH(Far)+ID_DEPTH(farLeaf)-2.0*node->Parent->depth >= maxDist)
        break;

      // Find the right sample size
      for (Size = Far->Size; Size; Size = Size->Next) {
	
	if (Size->SamSize == subSize) {

	  // When we find the right sample size, try all choices
	  // of number of nodes from the left subtree
	  for (Left = Size->Left; Left; Left = Left->Next) {
	    if (counting) {
	      *samples += Left->Samples;
	    } else {
	      *samples -= Left->Samples;
	      if (*samples < 0.0) {
		*NewCloseID = Close->ID;
		*NewFarID = Far->ID;
		*FromLeft = Left->FromLeft;
		return;
	      }
	    }
	  }
	
	  break;
	} else if (Size->SamSize > subSize) {
	  break;
	}
      }

      // If the far leaf was a descendent, ignore all other leaves
      if (isFarDesc)
	break;
    }

    // If the close leaf was a descendent, ignore all other leaves
    if (isCloseDesc)
      break;
  }

  ASSERT(counting, "Problem selecting a sample.");
}

// This function will return a list of leaves such that the sample
// satisfies the given constraints. The sample will consist of
// "subSize" leaves where no leaf is closer than "closeLeaf" and none
// is further than "farLeaf". If either of these leaves (or both) are 
// descendents of the given node, they are guaranteed to be included
// in the sample.
TreeNodeList * Pairwise_SelectSample(TreeNode * node, unsigned long closeLeaf,
			    unsigned long farLeaf, unsigned long subSize)
{
  double NumSamples, sampleIndex;
  unsigned long newCloseLeaf, newFarLeaf, FromLeft;
  TreeNodeList * Sample, * temp;
  
  ASSERT(node && closeLeaf <= numLeaves && farLeaf <= numLeaves &&
	 subSize > 0 && subSize <= sampleSize,
	 "Invalid parameter to SelectSample().");

  if (isLeaf(node)) {

    // If we are at a leaf then we better have been asking for
    // a sample of size one and this leaf better not be too
    // shallow or too deep. If so, report an error.
    ASSERT(subSize == 1 && node->depth >= LeafArray[closeLeaf]->depth &&
	   node->depth <= LeafArray[farLeaf]->depth,
	   "Sampling problems encountered at a leaf.");

    // This will be the one node in the sample
    Sample = (TreeNodeList *)malloc(sizeof(TreeNodeList));
    ASSERT(Sample, "Out of memory.");
    INC_MEM(sizeof(TreeNodeList));
    Sample->Node = node;
    Sample->Next = NULL;
    return Sample;
  }

  // Enumerate the number of valid samples satisfying the given
  // constraints rooted at this node
  NumSamples = -1.0;
  Pairwise_LoopThroughSamples(node, closeLeaf, farLeaf, subSize, &NumSamples, 
			      NULL, NULL, NULL);
  
  // If we don t find any valid samples here then something
  // went wrong with the sampling, so report an error.
  ASSERT(NumSamples > 0.0, "Got to a node with no valid samples.");

  // Generate a random index among all the potential samples
  sampleIndex = Pairwise_genrand() * NumSamples;

  // Find the correct choice of close leaf, far leaf, and
  // number of nodes to take from the left subtree based on
  // the randomly generated index
  Pairwise_LoopThroughSamples(node, closeLeaf, farLeaf, subSize, &sampleIndex, 
			      &newCloseLeaf, &newFarLeaf, &FromLeft);
  ASSERT(newCloseLeaf < numLeaves && newFarLeaf < numLeaves &&
	 FromLeft <= subSize,
	 "Problems selecting a sample.");

  // Ask the child nodes for the required samples
  if (FromLeft > 0) {

    // Get the right number of leaves from the left subtree
    Sample = Pairwise_SelectSample(node->LChild, newCloseLeaf, newFarLeaf, 
				   FromLeft);
    ASSERT(Sample, "Could not get a sample from the left child.");

    // If we didn t take all the leaves from the left subtree,
    // pick the rest from the right subtree
    if (FromLeft < subSize) {
      temp = Sample;
      while (temp->Next)
	temp = temp->Next;
      temp->Next = Pairwise_SelectSample(node->RChild, newCloseLeaf, 
					 newFarLeaf, subSize-FromLeft);
      ASSERT(temp->Next, "Could not get a sample from the right child.");
    }
  } else {

    // All the leaves should come from the right subtree, so
    // recursively sample from there
    Sample = Pairwise_SelectSample(node->RChild, newCloseLeaf, newFarLeaf, 
				   subSize);
    ASSERT(Sample, "Could not get a sample from the right child.");
  }

  return Sample;
}

