
#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include "structs.h"
#include "utils.h"

extern unsigned long sampleSize;
extern unsigned long numLeaves;
extern double CurMemoryUse, MaxMemoryUse;
extern TreeNode ** LeafArray;
static unsigned long CurLeafID = 0;

// Returns m! (only makes sense if m is really an integer)
double Pairwise_ffact(double m)
{
  double temp = 1.0;
  for (;m > 1.0; m--)
    temp = temp * m;
  return temp;
}

// Returns n choose m (only makes sense if m and n are integers)
double Pairwise_fchoose(double n, double m)
{
  if (m == 0 || m > n)
    return 0;
  else if (m == n)
    return 1;
 
  if (m > n/2.0) {
    return Pairwise_fchoose(n, n-m);
  } else {
    double i, temp = n / Pairwise_ffact(m);
    for (i=n-1.0; i > n-m; i--) {
      temp = temp * i;
    }
    return temp;
  }
}

// Returns the min of the two given values
double Pairwise_fmin(double x, double y)
{
  if (x <= y)
    return x;
  else
    return y;
}

// Returns the max of the two given values
double Pairwise_fmax(double x, double y)
{
  if (x >= y)
    return x;
  else
    return y;
}

// Traverses the tree updating the depth field at each node
void Pairwise_SetDepths(TreeNode * node, double depth)
{
  ASSERT(LeafArray, "SetDepths() called before LeafArray was created.");
  ASSERT(node, "SetDepths() called with a NULL node.");

  node->depth += depth;
  if (!isLeaf(node)) {
    Pairwise_SetDepths(node->LChild, node->depth);
    Pairwise_SetDepths(node->RChild, node->depth);
  } else {
    LeafArray[CurLeafID] = node;
    CurLeafID++;
  }
}

// Each leaf v has exactly one sample of size one rooted at v
// (where v is the closest and furthest node). Set this value
// for each leaf.
void Pairwise_PrepareLeaves()
{
  unsigned long i;

  ASSERT(LeafArray, "PrepareLeaves() called before LeafArray was created.");

  for (i=0; i < numLeaves; i++) {
    LeafArray[i]->Close = (CloseNode *)malloc(sizeof(CloseNode));
    ASSERT(LeafArray[i]->Close, "Out of memory.");
    INC_MEM(sizeof(CloseNode));

    LeafArray[i]->Close->ID = i;
    LeafArray[i]->Close->Next = NULL;
\
    LeafArray[i]->Close->Far = (FarNode *)malloc(sizeof(FarNode));
 ASSERT(LeafArray[i]->Close->Far, "Out of memory.");
    INC_MEM(sizeof(FarNode));

    LeafArray[i]->Close->Far->ID = i;
    LeafArray[i]->Close->Far->Next = NULL;

    LeafArray[i]->Close->Far->Size = (SizeNode *)malloc(sizeof(SizeNode));
    ASSERT(LeafArray[i]->Close->Far->Size, "Out of memory.");
    INC_MEM(sizeof(SizeNode));

    LeafArray[i]->Close->Far->Size->SamSize = 1;
    LeafArray[i]->Close->Far->Size->Next = NULL;

    LeafArray[i]->Close->Far->Size->Left = (LeftNode *)malloc(sizeof(LeftNode));
    ASSERT(LeafArray[i]->Close->Far->Size->Left, "Out of memory.");
    INC_MEM(sizeof(LeftNode));

    LeafArray[i]->Close->Far->Size->Left->Samples = 1.0;
    LeafArray[i]->Close->Far->Size->Left->FromLeft = 1;
    LeafArray[i]->Close->Far->Size->Left->Next = NULL;  
  }
}

// Uses selection sort to sort LeafArray by increasing depth and
// then assigns the correct ID to each leaf
void Pairwise_SortLeafArray()
{
  int i, j, min;
  TreeNode * temp;

  ASSERT(LeafArray, "SortLeafArray() called before LeafArray was created.");

  // Sort the leaves by depth
  for (i = 0; i < numLeaves-1; i++) {
    min = i;
    for (j=i+1; j < numLeaves; j++) {
      if (LeafArray[j]->depth < LeafArray[min]->depth)
	min = j;
    }
    temp = LeafArray[i];
    LeafArray[i] = LeafArray[min];
    LeafArray[min] = temp;
  }

  // Assign the correct ID to each leaf
  for (i=0; i < numLeaves; i++) {
    LeafArray[i]->ID = i;
  }
}

// Does the preprocessing of the tree. This basically amounts to
// assigning depths, creating the leaf array, and initializing
// the samples of size one at each of the leaf nodes.
void Pairwise_PrepareAllNodes(TreeNode * root)
{
  LeafArray = (TreeNode **)malloc(numLeaves * sizeof(TreeNode *));
  ASSERT(LeafArray, "Out of memory.");
  INC_MEM(numLeaves * sizeof(TreeNode *));

  Pairwise_SetDepths(root, 0.0);
  Pairwise_SortLeafArray();
  Pairwise_PrepareLeaves();
}

// Frees up any storage taken up by the tree rooted at the given node
void Pairwise_DeleteTree(TreeNode * node)
{
  unsigned long i,j,k;
  CloseNode * Close, * oldClose;
  FarNode * Far, * oldFar;
  SizeNode * Size, * oldSize;
  LeftNode * Left, * oldLeft;

  ASSERT(node, "DeleteTree() called with a NULL node.");

  // Delete the subtrees first
  if (!isLeaf(node)) {
    Pairwise_DeleteTree(node->LChild);
    Pairwise_DeleteTree(node->RChild);
  }

  // Delete the Samples matrix at this node
  Close = node->Close;
  while (Close) {
    Far = Close->Far;
    while (Far) {
      Size = Far->Size;
      while (Size) {
	Left = Size->Left;
	while (Left) {
	  oldLeft = Left;
	  Left = Left->Next;
	  free(oldLeft);
	}
	oldSize = Size;
	Size = Size->Next;
	free(oldSize);
      }
      oldFar = Far;
      Far = Far->Next;
      free(oldFar);
    }
    oldClose = Close;
    Close = Close->Next;
    free(oldClose);
  }

  free(node->Name);
  free(node);
}

// (Recursively) deletes a list of nodes
void Pairwise_DeleteList(TreeNodeList * list)
{
  ASSERT(list, "DeleteList() called with a NULL node.");
  if (list->Next)
    Pairwise_DeleteList(list->Next);
  free(list);
}

#define STATE_NONE                    0
#define STATE_READING_NAME_IN_QUOTES  1
#define STATE_READING_DIST            2
#define MAX_NAME_SIZE                 1024
#define SHORT_NAME_LENGTH             20

// Reads through the input file (in Newick format) and constructs a tree,
// returning a pointer to the root of the tree. The leaf nodes will
// have the name given to them in the file while the internal nodes
// will be named "Internal?" (where ? is a unique integer). The
// depth field at each node is actually its distance from its parent,
// not its real depth. Call SetDepths() after to fix this problem.
// If useShortNames is set to 1 then taxon names are truncated
// after 20 characters.
TreeNode * Pairwise_ReadInTree(FILE * fp, int useShortNames)
{
  TreeNode * curNode = NULL;
  TreeNode * rootNode = NULL;
  char * curName;
  char * curDist;
  int curNameIndex, curDistIndex;
  int wentUp = 0;
  int InternalIndex = 0;
  int curChar = 0, curLine = 0;
  int readingName = 0;
  char c, lastChar;
  int state = STATE_NONE;

  ASSERT(fp, "ReadInTree called with a NULL file.");

  curNode = NULL;

  // Don t bother counting these in the memory use because we free
  // them right after reading in the tree
  curName = malloc(MAX_NAME_SIZE+1);
  ASSERT(curName, "Out of memory.");
  curDist = malloc(255);
  ASSERT(curDist, "Out of memory.");

  curNameIndex = 0;
  curDistIndex = 0;

  numLeaves = 0;

  for (;;) {
    lastChar = c;
    c = fgetc(fp);
    if (feof(fp))
      break;

    if (c == '\n') {
      curChar = 0;
      curLine++;
      continue;
    } else {
      curChar++;
    }

    // File format is:
    //
    // T -> (T:dist,T:dist);
    // T -> name | 'name'
    //
    // Whitespace of any kind is ignored unless it
    // is part of a node name.

    if (c == '\'') {
      if (state == STATE_READING_NAME_IN_QUOTES) {
	state = STATE_NONE;
      } else {

	state = STATE_READING_NAME_IN_QUOTES;
	if (lastChar == '\'') {
	  curName[curNameIndex] = c;
	  curNameIndex++;
	}
      }
    } else if (state == STATE_READING_NAME_IN_QUOTES) {

      // Entering a name
      if ((!useShortNames && curNameIndex < MAX_NAME_SIZE-1)
	  || curNameIndex < SHORT_NAME_LENGTH) {
	curName[curNameIndex] = c;
	curNameIndex++;
      }
    } else if (isspace(c)) {
      continue;
    } else if (c == '(') {
      // New (internal) node. Create one and make it
      // the current node.
      TreeNode * newNode = malloc(sizeof(TreeNode));
      ASSERT(newNode, "Out of memory.");
      INC_MEM(sizeof(TreeNode));

      state = STATE_NONE;

#ifdef _DEBUG_
      // We don t really need to name the internal nodes, so only
      // do it if we are debugging.
      newNode->Name = (char *)malloc(MAX_NAME_SIZE * sizeof(char));
      ASSERT(newNode->Name, "Out of memory.");
      INC_MEM(MAX_NAME_SIZE * sizeof(char));
      sprintf(newNode->Name, "Internal%d", InternalIndex);
      InternalIndex++;
//++pac  #elsif
#else
      newNode->Name = NULL;
#endif

      newNode->LChild = NULL;
      newNode->RChild = NULL;
      newNode->depth = 0;
      newNode->ID = -1;
      newNode->Close = NULL;
      newNode->Parent = curNode;
      newNode->Descendents = NULL;

      // If this is the first node, make it the root. If not, find out
      // if it is a left child or a right child.
      if (rootNode) {
	if (newNode->Parent->LChild == NULL) {
	  newNode->Parent->LChild = newNode;
	} else {
	  newNode->Parent->RChild = newNode;
	}
      } else {
	rootNode = newNode;
      }

      curNode = newNode;

    } else if (c == ':') {
      state = STATE_READING_DIST;

      curDistIndex = 0;

      if (wentUp == 0) {

	// Finished entering a name. Create a node for
	// this element, and prepare to read distance
	// to parent.
	TreeNode * newNode = malloc(sizeof(TreeNode));
	ASSERT(newNode, "Out of memory.");
	INC_MEM(sizeof(TreeNode));

	curName[curNameIndex] = '\0';
	newNode->Name = (char *)strdup(curName);
	ASSERT(newNode->Name, "Out of memory.");
	INC_MEM((strlen(curName)+1) * sizeof(char));
	curNameIndex = 0;      
	newNode->LChild = NULL;
	newNode->RChild = NULL;
	newNode->Parent = curNode;
	newNode->Close = NULL;
	newNode->Descendents = NULL;

	if (curNode->LChild == NULL) {
	  curNode->LChild = newNode;
	} else {
	  curNode->RChild = newNode;
	}

	numLeaves++;
      }

      wentUp = 0;

      
    } else if (c == ',') {
      state = STATE_NONE;

      curDist[curDistIndex] = '\0';
      curNode->LChild->depth = atof(curDist);
      curDistIndex = 0;

    } else if (c == ')') {
      state = STATE_NONE;

      curDist[curDistIndex] = '\0';
      curNode->RChild->depth = atof(curDist);
      curDistIndex = 0;

      curNode = curNode->Parent;
      wentUp = 1;
    } else if (c == ';') {
      break;
    } else if (state != STATE_READING_DIST && isprint(c) && c != '[' && c != ']') {
      if ((!useShortNames && curNameIndex < MAX_NAME_SIZE-1)
	  || curNameIndex < SHORT_NAME_LENGTH) {
	curName[curNameIndex] = c;
	curNameIndex++;
      }
    } else if (state == STATE_READING_DIST && 
	       (isdigit(c) || c == '.' || c == 'e' || c == '-')) {
      
      // Entering a distance
      curDist[curDistIndex] = c;
      curDistIndex++;
    } else {

      free(curName);
      free(curDist);
      ASSERT3(0, "Something is wrong with the format of the input file.\nThe error was detected at line %d, character %d.", curLine, curChar);
    }
  }

  free(curName);
  free(curDist);

  return rootNode;
}
